/* $id: sqlora.c,v 1.48 2002/07/19 12:37:24 kpoitschke Exp $ */
/**
 * @file sqlora.c
 * libsqlora8 Implementation
 *
 * @author Kai Poitschke
 *
 * Copyright (c) 1991-2002 Kai Poitschke (kai[_at]poitschke.de)
 *
 *   This file is part of the libsqlora8 package which can be found
 *   at http://www.poitschke.de/libsqlora8/
 */

/*
 *     Permission to use, copy, modify, and distribute this software for
 *     any purpose with or without fee is hereby granted, provided that
 *     the above copyright notice and this permission notice appear in all
 *     copies.
 *
 *     THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *     WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 *     MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *     IN NO EVENT SHALL THE AUTHORS AND COPYRIGHT HOLDERS AND THEIR
 *     CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *     SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *     LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *     USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *     ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *     OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *     OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *     SUCH DAMAGE.
 */

const char * _sqlo_sqloraID="$Id: sqlora.c,v 1.51.2.4 2002/12/06 16:13:53 kpoitschke Exp $";

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <errno.h>

#ifdef HAVE_UNISTD_H            /* for my windows friends  */
#    include <unistd.h>
#endif
#include <ctype.h>
#include <assert.h>

#if defined(HAVE_PTHREAD_H)
#include <pthread.h>
#endif

#include "oci.h"
#include "sqlora.h"

/* If glib is enabled, we use the glib allocators */
#ifdef USE_GLIB_ALLOC
#  include "glib.h"
#  define MALLOC g_malloc
#  define REALLOC g_realloc
#  define FREE g_free
#else
#  define MALLOC malloc
#  define REALLOC realloc
#  define FREE free 
#endif

#if defined(WIN32)
#define inline __inline
#endif

#if defined (__STDC__) || defined (_AIX) || defined(PROTOTYPES) ||\
            (defined (__mips) && defined (_SYSTYPE_SVR4)) ||\
             defined(WIN32) || defined(__cplusplus)\
            
#  define AND ,
#  define DEFUN(name, arglist, args)      name(args)
#  define DEFUN_VOID(name)                name(void) 
/*
 * Macro to use instead of "void" for arguments that must have
 * type "void *" in ANSI C;  maps them to type "char *" in
 * non-ANSI systems.
 */
#  define VOID void
#else
#  define AND ;
#  define DEFUN(name, arglist, args)      name arglist args;
#  define DEFUN_VOID(name)                name()
#  define VOID char
#endif

#ifndef NULL
#   define NULL 0
#endif

#ifdef TEST_WITH_SIZEOF_RETURNING_ULONG
#define sizeof(_a) (unsigned long)sizeof(_a)
#endif

#if defined(ENABLE_PTHREADS) || defined(ENABLE_ORATHREADS)
/**
 * @enum
 * Defines if the library was compiled with --enable-pthreads
 */
enum {THREADS_ENABLED = 1};
#else
enum {THREADS_ENABLED = 0};
#endif

#ifdef ENABLE_PTHREADS
typedef pthread_t sqlo_thread_t;
typedef pthread_mutex_t sqlo_mutex_t;
#elif ENABLE_ORATHREADS
typedef OCIThreadMutex * sqlo_mutex_t;
typedef OCIThreadId * sqlo_thread_t;
#else
typedef unsigned long sqlo_mutex_t; /* dummy */
typedef unsigned long sqlo_thread_t; /* dummy */
#endif

/**
 * @def EXEC_WHEN_THREADING
 * Executes _cmd if library was compiled with threading and initialized in threaded mode
 * @param _cmd I - The code to execute
 */
#define EXEC_WHEN_THREADING(_cmd) \
     if (THREADS_ENABLED && OCI_THREADED == _oci_init_mode) { _cmd }

/**
 * @def UNLOCK_ALL
 * Release all mutex locks
 */
#define UNLOCK_ALL EXEC_WHEN_THREADING(_dbv_unlock(); _stmtv_unlock(); _env_unlock(); )


/*-------------------------------------------------------------------------
 * CONSTANTS
 *-----------------------------------------------------------------------*/
/**
 * @enum _sqlora_constants
 */
enum _sqlora_constants {                       

  MIN_BINDP = 64,               /**< Initial number of bind pointers
                                 * that were allocated 
                                 */

  MIN_DEFNP = 32,               /**< Initial number of define pointers
                                 * that were allocated 
                                 */

  MIN_OBUF_SIZE = 32,           /**< Minimum size of the output buffer for columns 
                                 * in the select list 
                                 */

  MIN_COL_NAME_LEN = 31,        /**< Minimum size for a name or a column */

  MIN_STMT_SIZE =  1024,        /**< Mininum allocation for sqlo_stmt_t.stmt */

  DEF_PREFETCH_ROWS = 100,      /**< The default number of prefetched rows */

  SQLO_MAX_ERRMSG_LEN = 2047,   /**< The maximum length of the error message buffer */

  MAX_PATH_LEN = 512,           /**< for the trace file */
  
  MAX_VNAME_LEN = 255,          /**< Max variable name len read from the environment */
  
  MAX_LONG_SIZE = (1024*64),     /**< Max size for LONG output variables */

#ifdef LIBSQLORA8_TRACE_ENABLED
  TRACE_ENABLED = 1      /**< configured with enabled trace */
#else
  TRACE_ENABLED = 0      /**< configured with disabled trace */
#endif

};


/*-------------------------------------------------------------------------
 * MACROS
 *-----------------------------------------------------------------------*/
#undef TRUE
#undef FALSE

/**
 * Boolean TRUE
 */
#define TRUE ((1 == 1))

/**
 * Boolean FALSE
 */
#define FALSE (!TRUE)

/**
 * @def TRACE
 * Execute cmd when the trace level of the library is >= the trace level
 * passed to the macro.
 * @param _trc_level   I - The min trace level where cmd should be executed
 * @param _cmd         I - The commands to be executed in case of 
 *                         _trace_level >= trc_level.
 *
 * @par Example:
 * TRACE(3, fprintf(g_ftp, "Calling foo()\n"););
 */
#define TRACE(_trc_level, _cmd) if (TRACE_ENABLED) \
{\
   if ( (NULL != _trace_fp) && _trace_level >= _trc_level) {\
      { _cmd } \
      (void) fflush(_trace_fp); \
   }\
}

/**
 * @def ERRMALLOC
 * Allocates _len bytes memory for _ptr
 *
 * If no memory could be allocated, an error message is
 * put in _dbp->errmsg and the macro returns _retcode
 * _func should be a string identifying the
 * location (function name) where the ERRMALLOC was called
 * @param _dbp     I - The pointer to the database strucuture
 * @param _ptr     O - The allocated memory
 * @param _len     I - The number of bytes to allocate
 * @param _func    I - The function name where this macro was called
 * @param _retcode I - The return code, the caller wants to return if the malloc fails.
 */
#define ERRMALLOC(_dbp, _ptr, _len, _func, _retcode)   \
{\
   (_ptr) = MALLOC((_len));\
   if (!(_ptr) && (_dbp)) {\
	sprintf((_dbp)->errmsg,\
            "Cannot malloc %u bytes in %s\n", (unsigned int)(_len), _func);\
	TRACE(1, (void) fputs((_dbp)->errmsg, _get_trace_fp(_dbp)););\
    /* make sure we release all locks */ \
    UNLOCK_ALL \
	return _retcode;\
   }\
   TRACE(4, fprintf(_get_trace_fp(_dbp), "Allocated %u bytes in %s\n", \
                    (unsigned int)(_len), _func););\
}

/**
 * @def ERREALLOC
 * Rellocates _len bytes memory at _ptr.
 * If no memory could be allocated, an error message is
 * put in _dbp->errmsg and the macro returns _retcode
 * _func should be a string identifying the
 * location (function name) where the ERRMALLOC was called
 * @param _dbp      I - The pointer to the database strucuture
 * @param _ptr      O - The allocated memory
 * @param _len      I - The number of bytes to allocate
 * @param _func     I - The function name where this macro was called
 * @param _retcode  I - The return code, the caller wants to return if the malloc fails.
 */
#define ERREALLOC(_dbp, _ptr, _len, _func, _retcode)   \
{\
   (_ptr) = REALLOC((_ptr), (_len));\
   if (!(_ptr) && (_dbp)) {\
	sprintf((_dbp)->errmsg,\
            "Cannot realloc %u bytes in %s\n", (unsigned int)(_len), _func);\
	TRACE(1, (void) fputs((_dbp)->errmsg, _get_trace_fp(_dbp)););\
    /* make sure we release all locks */ \
    UNLOCK_ALL \
	return _retcode;\
   }\
   TRACE(4, fprintf(_get_trace_fp(_dbp), "Reallocated %u bytes in %s\n", \
                                 (unsigned int)(_len), _func););\
}

/**
 * @def CHECK_DBHANDLE
 * Checks for a valid dbh and sets the dbp or returns with error.
 * If the db handle is invalid, the macro calls return _errval.
 * @param _dbp    O - The pointer to the _sqlo_db_t structure for the _dbh.
 * @param _dbh    I - The database handle to check.
 * @param _func   I - The callers function name. Used for error message.
 * @param _errval I - The value to return in case of an error.
 */
#define CHECK_DBHANDLE(_dbp, _dbh, _func, _errval)\
{\
  if (_dbh >= (int) _dbv_size || _dbh < 0 || !_dbv[ _dbh ] || !_dbv[ _dbh ]->used) {\
    TRACE(1, fprintf(_trace_fp, "Invalid Database handle %d in %s\n",\
                      _dbh, _func);); \
    /* make sure we release all locks */ \
    UNLOCK_ALL; \
    return (_errval); \
  }\
  _dbp = _dbv[_dbh];\
}

/**
 * @def CHECK_STHANDLE
 * Checks for a valid sth and sets the stp or returns with error
 * If the db handle is invalid, the macro calls return _errval.
 * @param _stp    O - The pointer to the _sqlo_stmt_t structure for the _sth.
 * @param _sth    I - The statement handle to check.
 * @param _func   I - The callers function name. Used for error message.
 * @param _errval I - The value to return in case of an error.
 */
#define CHECK_STHANDLE(_stp, _sth, _func, _errval) \
{\
  if (_sth >= (int) _stmtv_size || _sth < 0 || !_stmtv[ _sth ] || !_stmtv[ _sth ]->used) {\
      sqlo_db_ptr_t  xdbp; \
      if ((xdbp = (_dbv[0]))) { \
          sprintf (xdbp->errmsg, "Invalid sth %d passed to %s (line: %d)\n",\
                                 _sth, _func,__LINE__); \
      }\
      TRACE(1, fprintf(_trace_fp, "Invalid sth %d passed to %s (line: %d)\n",\
                              _sth, _func, __LINE__););\
      /* make sure we release all locks */ \
      UNLOCK_ALL; \
      return (_errval);\
    }\
  _stp = _stmtv[_sth];\
}

/**
 * @def CHECK_OCI_STATUS
 * Checks and saves the OCI return status.
 * If the status is != OCI_SUCCESS, the function save_status is called to
 * save the status in the database structure.
 *
 * @param _dbp    I - The pointer to the database structure.
 * @param _stat   I - The status to check.
 * @param _action I - The action you executed causing this status
 * @param _object I - The object on which you did _action.
 */
#define CHECK_OCI_STATUS(_dbp, _stat, _action, _object) { \
        (_dbp)->status = _stat;\
        TRACE(4, fprintf(_get_trace_fp(_dbp), "CHECK_OCI_STATUS[%u]: %d at %d\n", \
                         _dbp->dbh, _stat, __LINE__););\
        if (OCI_SUCCESS != _stat) {\
          _save_oci_status(_dbp, _action, _object, __LINE__);\
        }\
}

/**
 * @def CHECK_OCI_STATUS_RETURN
 * Checks and saves the OCI status and returns the status on failure.
 *
 * Calls CHECK_OCI_STATUS and returns the status if it is != OCI_SUCCESS.
 *
 * @param _dbp  I - The pointer to the database structure.
 * @param _stat I - The status to check.
 * @param _action I - The action you executed causing this status
 * @param _object I - The object on which you did _action.
 */
#define CHECK_OCI_STATUS_RETURN(_dbp, _stat, _action, _object) { \
        CHECK_OCI_STATUS(_dbp, _stat, _action, _object);\
        if (OCI_SUCCESS != _stat) { \
          UNLOCK_ALL; \
          return (_stat); \
        }\
}


/* If we have usleep we wait this amount of microseconds in a 
 * OCI_STILL_EXECUTING loop
 */
#ifdef HAVE_USLEEP
#  define SQLO_USLEEP usleep(1000)
#else
#  define SQLO_USLEEP
#endif


/*-------------------------------------------------------------------------
 * TYPES
 *-----------------------------------------------------------------------*/

/**
 * Boolean type
 */
typedef unsigned char bool_t;


/**
 * Cursor types
 */
typedef enum {
  DEFAULT   = 0,                /**< standard cursor */
  REFCURSOR = 1,                /**< ref cursor type */
  NTABLE    = 2                 /**< nested table cursor */
} sqlo_cursor_type_e;

/**
 * Stores information about the database connection.
 */
typedef struct {

  unsigned int dbh;              /**< The db handle used to address this entry */


  OCIServer * srvhp;                                     /**< OCI server handle */

  OCIError * errhp;                                       /**< OCI error handle */

  OCISvcCtx * svchp;                            /**< OCI service context handle */

  OCISession * authp;                     /**< OCI authorization session handle */

  char * tnsname;                             /**< The TNS name of the database */

  int status;                                        /**< The last status code */

  char errmsg[2*SQLO_MAX_ERRMSG_LEN+1];              /**< The last error message */

  sb4 errcode;                                 /**< The last oracle error code */

  OCIEnv * envhp;               /**< pointer to the OCI environment for the thread */

  OCIStmt * stmthp;               /**< @ref sqlo_exec stores its stmthp here.
                                     This is not dangerous, because during a non-blocking
                                     OCI call, you cannot open other stmts 
                                  */

  /* CONTROL FLAGS */
  bool_t used;                    /**< Flag, if this one is in use or not */
  bool_t attached;                /**<  not zero, if we are attached to the server */
  bool_t session_created;         /**< not zero, if a session was created */

  FILE * trace_fp;              /**< connection specific trace file */

} sqlo_db_t, * sqlo_db_ptr_t;

/**
 * This pointer type defines a pointer to a const sqlo_db_t.
 */
typedef const sqlo_db_t * const_sqlo_db_ptr_t;

struct _sqlo_stmt_t;

/**
 * Stores information about a column in the select list.
 * This structure is not used if you bind the select list manually be
 * @ref sqlo_define_by_pos or @ref sqlo_define_by_pos2.
 */
typedef struct  {

  unsigned pos;                /**< The position in the select list (0 based).
   * This variable can be used to address the right
   * data or ind element in the stmt structure. */

  OCIParam * paramd;           /**< The parameter handle pointer used to describe
   * an output column
   */
  ub2  dtype;                   /**< The datatype (@ref sqlo_data_types) */

  char * col_name;              /**< The column name */

  unsigned col_name_size;       /**< The max allocated length of col_name */

  ub2  dbsize;                  /**< The size in the database */

  ub1  prec;                    /**< The precision */

  ub1  scale;                   /** The scale */

  ub1  nullok;                  /**< Flag: Null allowed */

  struct _sqlo_stmt_t * stp;    /**< link to the stmt structure (see @ref sqlo_stmt_t) */

} sqlo_col_t, * sqlo_col_ptr_t;

/**
 * Stores information about an sql statement.
 * This structure stores all OCI handles plus the data buffers used to
 * retrieve the data.
 */
typedef struct {
  unsigned sth;                     /**< The own handle that identifies this entry */

  sqlo_db_ptr_t  dbp;               /**< The link to the database connection (see @ref sqlo_db_t). */ 

  OCIStmt * stmthp;              /**< The OCI statement handle pointer. */


  char * stmt;                   /**< The sql statement. */

  unsigned stmt_size;           /**< The allocated size of  stmt. */

  /* INPUT */
  OCIBind ** bindpv;             /**< The vector of input bind variables. */

  unsigned num_bindpv;          /**< The number of used entries in bindpv[]. */

  unsigned bindpv_size;         /**< The current capacity of size of bindpv[] and indp[]. */

  ub2 stype;                    /**< The OCI Statement type (see oci.h). */

  short * indpv;                /**< Indicator variable vector for input bind variables. */
  
  /* OUTPUT */
  sqlo_col_t * ocolsv;           /**< The output columns of the select list. */

  OCIDefine ** defnpv;           /**< The OCI define pointers. */

  unsigned num_defnpv;          /**< The current number of output variables. */

  unsigned defnpv_size;         /**< The current size of the arrays defnpv, outv, 
                                   outv_size,  oindv, rlenv, ocol_namev_size and 
                                   ocol_namev_size.
                                */

  char ** outv;                 /**< The output columns as a vector. 
                                   See @ref sqlo_values. 
                                */

  unsigned * outv_size;         /**< The current size of an outv[i] element */

  ub2 * oindv;                  /**< The indicator array for output variables */

  ub2 * rlenv;                  /**< The  actual length of a returned column. */

  char ** ocol_namev;           /**< Output column names. 
                                   The actual number of filled entries is num_defnpv#.
                                */

  unsigned * ocol_namev_size;        /**< The sizes of a colum name  in ocol_namev[]. */

  sqlo_cursor_type_e cursor_type; /**< The cursor type see @ref sqlo_cursor_type_e */

  /* status flags */
  bool_t opened;                            /**< 1 if the cursor is open, 0 if not. */ 

  bool_t prepared;                   /**< 1 if the statement is prepared, 0 if not. */

  bool_t used;             /**< 1 indicates that this structure is in use, 0 if not. */

  bool_t still_executing;  /**< Is 1 in non-blocking mode and a call 
                              to OCIStmtExecute returned OCI_STILL_EXECUTING 
                           */

  bool_t num_executions;       /**< number of times the statement was executed */

  sqlo_thread_t thread_id; /**< The thread id of the thread who opened this cursor */

} sqlo_stmt_t, * sqlo_stmt_ptr_t;

/**
 * This pointer type defines a pointer to a const sqlo_stmt_t.
 */
typedef const sqlo_stmt_t * const_sqlo_stmt_ptr_t;

/**
 * Structure to store parameter name, pointer to the variable
 * and a optional trigger fct, that is executed when parameter is 
 * changed. 
 * @see g_params.
 */
typedef struct {
  char * name;                                          /**< parameter name */

  enum {INTEGER, STRING} vtyp;                           /**< type of value */

  VOID * value;                               /**< The address of the value */

  int (*trigger_fct) __P((int));       /**< Function that handles this type */

} sqlora_param_t;

/*-------------------------------------------------------------------------
 * EXPORTED GLOBALS
 *-----------------------------------------------------------------------*/

/* define some variables where the user can check the version */
const unsigned sqlo_major_version = LIBSQLORA8_MAJOR_VERSION;
const unsigned sqlo_minor_version = LIBSQLORA8_MINOR_VERSION;
const unsigned sqlo_micro_version = LIBSQLORA8_MICRO_VERSION;
const unsigned sqlo_interface_age = LIBSQLORA8_INTERFACE_AGE;
const unsigned sqlo_binary_age    = LIBSQLORA8_BINARY_AGE;

/* for backward compatibility */
const unsigned sqlora8_major_version = LIBSQLORA8_MAJOR_VERSION;
const unsigned sqlora8_minor_version = LIBSQLORA8_MINOR_VERSION;
const unsigned sqlora8_micro_version = LIBSQLORA8_MICRO_VERSION;
const unsigned sqlora8_interface_age = LIBSQLORA8_INTERFACE_AGE;
const unsigned sqlora8_binary_age    = LIBSQLORA8_BINARY_AGE;


/*-------------------------------------------------------------------------
 * MODULE GLOBAL VARIABLES
 *-----------------------------------------------------------------------*/

/**
 * @var _errmsg
 * A buffer for the error messages.
 * This buffer is used when the connection specific buffer cannot be used,
 * because there is no db structure allocated.
 */
static char _errmsg[SQLO_MAX_ERRMSG_LEN+1];

/**
 * @var _oci_init_mode
 * The mode of the library.
 * Either OCI_DEFAULT or OCI_THREADED.
 */
static int _oci_init_mode = OCI_DEFAULT; /* the mode we initialize the OCI lib */


/**
 * @var _max_long_size
 *
 * The maxiumum length of a LONG result.
 * Can be changed by setting the environment variable SQLORA_LONGSIZE.
 * Default is @ref MAX_LONG_SIZE.
 */
static unsigned int _max_long_size = MAX_LONG_SIZE; 

/**
 * @var _dbv
 * The array of @ref sqlo_db_t pointers.
 * The array is allocated in @ref sqlo_init with the size of the passed max_db
 * parameter. 
 * The array is proteced by the mutex @ref _dbv_mux when compiled with threading
 * enabled.
 */
static sqlo_db_t ** _dbv = NULL; /* array for database connections */

/**
 * @var _dbv_size
 * The size of the @ref _dbv[].
 */
static unsigned int _dbv_size = 0;           /* number of available entries in _dbv[] */

/**
 * @var _env_mux
 * A mutex to protect the call to OCIEnvCreate
 */
#if ENABLE_PTHREADS
static pthread_mutex_t _env_mux = PTHREAD_MUTEX_INITIALIZER;
#elif ENABLE_ORATHREADS
static OCIThreadMutex *_env_mux;
#endif

/**
 * @var _dbv_mux
 * A mutex to protect @ref _dbv.
 * @see _dbv_lock, _dbv_unlock
 */
#if ENABLE_PTHREADS
static pthread_mutex_t _dbv_mux = PTHREAD_MUTEX_INITIALIZER;
#elif ENABLE_ORATHREADS
static OCIThreadMutex *_dbv_mux; 
#endif

/**
 * @var _stmtv_mux
 * A mutex to protect the array @ref _stmtv when compiled in threaded mode.
 */
#if ENABLE_PTHREADS
static pthread_mutex_t _stmtv_mux = PTHREAD_MUTEX_INITIALIZER;
#elif ENABLE_ORATHREADS
static OCIThreadMutex *_stmtv_mux;
#endif

/**
 * @var _stmtv
 * A array to store all @ref sqlo_stmt_t pointers.
 * This array is allocated in @ref sqlo_init with the size of
 * the parameter max_cursors.
 * The array is proteced by the mutex @ref _stmtv_mux when compiled in threaded mode.
 */
static sqlo_stmt_t **_stmtv = NULL; /* array to hold statements */

/**
 * @var _stmtv_size
 * The max size of @ref _stmtv.
 */
static unsigned int _stmtv_size = 0;             /* max size of _stmtv[] */



/**
 * @var _num_prefetch_rows
 * The number of rows Oracle should prefetch for us.
 * The variable is initialized with @ref DEF_PREFETCH_ROWS, but can
 * be changed by the environment variable SQLORA_PREFETCH_ROWS.
 */
static ub4 _num_prefetch_rows = DEF_PREFETCH_ROWS; /* out prefetch value */

/**
 * @var _sqlo_init
 * A flag indicating if the library was successfully initialized
 */
static int _sqlo_init = 0;                      /* Is set to 1 by sqlo_init */

/**
 * @var _trace_level
 *
 * The trace level of the library.
 * The level is initialized from the environment variable SQLORA_TRACE_LEVEL
 *
 * <ul>
 * <li>0 : Trace disabled (default)
 * <li>1 : Print errors to tracefile
 * <li>2 : Print function calls to tracefile
 * <li>3 : Print more detailed information to tracefile
 * <li>4 : Print also malloc/realloc operations.
 * </ul>
 */
static int _trace_level = 0; 

/**
 * @var _trace_file
 * The name of the trace file.
 * Default is sqlora8.trc, but can be changed by setting the environment
 * variable SQLORA_TRACE_FILE
 */
static char _trace_file[MAX_PATH_LEN+1] = "";

/* These handles are needed by the Oracle OCIThread package.
 * We cannot use the ones in _dbv, because this is the structure
 * we want to protect.
 */
#ifdef ENABLE_ORATHREADS
static OCIEnv *_oci_envhp;
static OCIError *_oci_errhp;
#endif


/**
 * @var DEFAULT_TRACE_FNAME
 * The default trace filename.
 */
static const char * DEFAULT_TRACE_FNAME = "sqlora8.trc"; 

static FILE * _trace_fp = NULL;     /**< The filepointer of the global tracefile  */

static unsigned int _session_count = 0; /**< The nubmer of created sessions. Used to name the trace file */

/*---------------------------------------------------------------------------
 * PROTOTYPES
 *--------------------------------------------------------------------------*/
/* functions needed in threaded mode */
static int _init_mutexes __P((void));
static int _dbv_lock __P((void));
static int _dbv_unlock __P((void));
static int _stmtv_lock __P((void));
static int _stmtv_unlock __P((void));
static int _env_lock __P((void));
static int _env_unlock __P((void));

static int _sqlo_getenv __P((void));

static int _save_oci_status __P((sqlo_db_ptr_t dbp, 
                                 const char *action, 
                                 const char *object, 
                                 int lineno));

static int _bind_argv __P(( sqlo_stmt_ptr_t  stp, unsigned int argc, const char ** argv));

static int _bind_by_pos __P((sqlo_stmt_ptr_t  stp, unsigned int param_pos, int param_type,
                            const void * param_addr, unsigned int  param_size,
                            short * ind_addr, int is_array));

static int _bind_by_pos2 __P((sqlo_stmt_ptr_t  stp, unsigned int param_pos, int param_type,
                            const void * param_addr, unsigned int  param_size,
                            short * ind_addr, unsigned short * rcode_addr, 
                               unsigned int skip_size));

static void _strip_string __P((char *s, unsigned int len));

static int _define_ocol_by_pos __P((sqlo_stmt_ptr_t stp, sqlo_col_t *colp, 
                                    unsigned int pos));

static int _define_output __P((sqlo_stmt_ptr_t stp));
static int _open_trace_file __P((sqlo_db_ptr_t dbp));
static sqlo_stmt_ptr_t _get_stmt_ptr __P((const_sqlo_db_ptr_t dbp));
static int _stmt_new __P((sqlo_db_ptr_t dbp, const char * stmt, sqlo_stmt_ptr_t *stpp));
static void _stmt_release __P((sqlo_stmt_ptr_t  stp));
static void _bindpv_reset __P((sqlo_stmt_ptr_t  stp));
static int _stmt_init __P((sqlo_stmt_ptr_t  stp, sqlo_db_ptr_t dbp, const char *stmt));

static const char * _get_stmt_type_str __P((int stype));
static const char * _get_data_type_str __P((int dtype));
static sqlo_db_ptr_t _db_add __P((void));
static void _db_release __P((sqlo_db_ptr_t dbp));

static int _define_by_pos __P((sqlo_stmt_ptr_t  stp, unsigned int value_pos, 
                               int value_type, const void * value_addr, 
                               unsigned int  value_size, short * ind_addr, 
                               ub2 * rlen_addr, ub2 * rcode_addr, int is_array));

static int _define_by_pos2 __P((sqlo_stmt_ptr_t  stp, unsigned int value_pos, 
                                int value_type, const void * value_addr, 
                                unsigned int  value_size, short * ind_addr, 
                                ub2 * rlen_addr, ub2 * rcode_addr, 
                                unsigned int skip_size));

static  int _calc_obuf_size __P(( unsigned int *bufsizep, unsigned int data_type,
                                  int prec, int scale, unsigned int dbsize));

static int _get_blocking_mode __P((sqlo_db_ptr_t dbp, unsigned * blocking));

static int _get_errcode __P(( sqlo_db_ptr_t dbp));

static int _set_prefetch_rows __P((sqlo_stmt_ptr_t stp, unsigned int nrows));

static const char * _get_stmt_string __P((sqlo_stmt_ptr_t stp));
static int _get_stmt_state __P((sqlo_stmt_ptr_t stp));
static int _alloc_definep __P((sqlo_stmt_ptr_t stp, unsigned int size));
static int _alloc_bindp __P((sqlo_stmt_ptr_t stp, unsigned int size));
static void _close_all_db_cursors __P((const_sqlo_db_ptr_t dbp));
static void _close_all_executing_cursors __P((const_sqlo_db_ptr_t dbp));
static FILE * _get_trace_fp __P((const_sqlo_db_ptr_t dbp));
static int _prepare __P((const_sqlo_stmt_ptr_t stp, const char * stmt, ub2 * stmt_type));
static sqlo_thread_t _get_thread_id __P((void));
static bool_t _thread_id_equal __P(( sqlo_thread_t id1, sqlo_thread_t id2));

/* Prototypes of functions not present in oracle header files */
int osnsui __P((int * handlep, sqlo_signal_handler_t signal_handler, char * ctx));
int osncui __P((int handle));

/*---------------------------------------------------------------------------
 * END PROTOTYPES
 *--------------------------------------------------------------------------*/


#ifdef HAVE_STRDUP
/* Sometimes strdup is  not defined in strict ansi mode,
 * even if it is in libc 
 */
extern char * strdup __P((const char *s));
#else

/**
 * strdup if not available on the target system
 * @param s  I - The string to duplicate
 * @return The newly allocated string.
 */
char * DEFUN(strdup, (s), const char * s)
{
  char *n;
  n = (char *) MALLOC(sizeof(char) * (strlen(s) + 1) );
  TRACE(4, fprintf(_trace_fp,"strdup: Allocated %d bytes\n", (strlen(s) + 1)););
  if (n)
    strcpy(n, s);
  return(n);
}

#endif



/*---------------------------------------------------------------------------
  Parameters (Can be changed by environment variables or sqlo_set)
  Naming conventions of environment variables:
  SQLORA_ || upper(<param_name>)
---------------------------------------------------------------------------*/
          /* Parameters, sort them by name */
static sqlora_param_t g_params[] = {
  {"PREFETCH_ROWS", INTEGER,       (VOID *)&_num_prefetch_rows,    NULL},
  {"LONGSIZE",      INTEGER,       (VOID *)&_max_long_size,    NULL},
  {"TRACE_FILE",    STRING,        _trace_file,          NULL},
  {"TRACE_LEVEL",   INTEGER,       &_trace_level,        NULL},
  {NULL,            INTEGER,       NULL,                NULL}
};

/*-------------------------------------------------------------------------
 * STATIC FUNCTIONS
 *-----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------*/
/**
 * Inits all mutexes.
 * @returns OCI status.
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _init_mutexes
#endif
static inline int
DEFUN_VOID(_init_mutexes)
{
  int status = SQLO_SUCCESS;
#ifdef ENABLE_PTHREADS
  pthread_mutex_init(&_dbv_mux, NULL);
  pthread_mutex_init(&_stmtv_mux, NULL);
  pthread_mutex_init(&_env_mux, NULL);

#elif ENABLE_ORATHREADS
  status = (OCIThreadMutexInit(_oci_envhp, _oci_errhp, &_dbv_mux) ||
            OCIThreadMutexInit(_oci_envhp, _oci_errhp, &_stmtv_mux) ||
            OCIThreadMutexInit(_oci_envhp, _oci_errhp, &_env_mux) );
#endif
  return status;
}

/*-------------------------------------------------------------------------*/
/**
 * Lock @ref _env_mux
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _env_lock
#endif
static inline int
DEFUN_VOID(_env_lock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_lock(&_env_mux);
#elif ENABLE_ORATHREADS
  return(OCIThreadMutexAcquire(_oci_envhp, _oci_errhp, _env_mux));
#endif
  return OCI_SUCCESS;
}

/*-------------------------------------------------------------------------*/
/**
 * Unlock @ref _env_mux.
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _env_unlock
#endif
static inline int
DEFUN_VOID(_env_unlock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_unlock(&_env_mux);
#elif ENABLE_ORATHREADS
  return (OCIThreadMutexRelease(_oci_envhp, _oci_errhp, _env_mux));
#endif 
  return(OCI_SUCCESS);
}

/*-------------------------------------------------------------------------*/
/**
 * Lock @ref _dbv
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _dbv_lock
#endif
static inline int
DEFUN_VOID(_dbv_lock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_lock(&_dbv_mux);
#elif ENABLE_ORATHREADS
  return(OCIThreadMutexAcquire(_oci_envhp, _oci_errhp, _dbv_mux));
#endif
  return OCI_SUCCESS;
}

/*-------------------------------------------------------------------------*/
/**
 * Unlock @ref _dbv.
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _dbv_unlock
#endif
static inline int
DEFUN_VOID(_dbv_unlock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_unlock(&_dbv_mux);
#elif ENABLE_ORATHREADS
  return (OCIThreadMutexRelease(_oci_envhp, _oci_errhp, _dbv_mux));
#endif 
  return(OCI_SUCCESS);
}


/*-------------------------------------------------------------------------*/
/**
 * Lock @ref _stmtv
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _stmtv_lock
#endif
static inline int
DEFUN_VOID(_stmtv_lock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_lock(&_stmtv_mux);
#elif ENABLE_ORATHREADS
  return (OCIThreadMutexAcquire(_oci_envhp, _oci_errhp, _stmtv_mux));
#endif
  return(OCI_SUCCESS);
}


/*-------------------------------------------------------------------------*/
/**
 * Unlock @ref _stmtv.
 * @return OCI status
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _stmtv_unlock
#endif
static inline int
DEFUN_VOID(_stmtv_unlock)
{
#ifdef ENABLE_PTHREADS
  pthread_mutex_unlock(&_stmtv_mux);
#elif ENABLE_ORATHREADS
  return (OCIThreadMutexRelease(_oci_envhp, _oci_errhp, _stmtv_mux));
#endif
  return(OCI_SUCCESS);
}



/*-------------------------------------------------------------------------*/
/**
 * Sets the attribute PREFETCH_ROWS.
 * @param stp   I - statement pointer
 * @param nrows I - Number of rows to prefetch
 * @return OCI status code
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _set_prefetch_rows
#endif
static inline int
DEFUN(_set_prefetch_rows, (stp, nrows), 
      sqlo_stmt_ptr_t  stp AND
      unsigned int nrows)
{
  unsigned int prefetch_rows = nrows;
  int status;

  status = OCIAttrSet((dvoid*)stp->stmthp, 
                        (ub4)OCI_HTYPE_STMT,
                        &prefetch_rows, 
                        (ub4) sizeof(prefetch_rows),
                        (ub4)OCI_ATTR_PREFETCH_ROWS, 
                        stp->dbp->errhp);
  return(status);
}


/*-------------------------------------------------------------------------*/
/**
 * Gets the statement text and puts it into stp->stmt.
 *
 * @param stp   I - statement pointer
 * @return OCI status code
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _get_stmt_string
#endif
static inline const char *
DEFUN(_get_stmt_string, (stp),
      sqlo_stmt_ptr_t  stp 
     )
{
  static const char *nostmt = "_get_stmt_string: No statement avalailable";
  if (!stp->stmt)
    return(nostmt);
  
  return stp->stmt;
}


/*-------------------------------------------------------------------------*/
/**
 * Gets the statement state
 *
 * @param stp   I - statement pointer
 * @return The statement state or -1 on error.
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _get_stmt_state
#endif
static inline int
DEFUN(_get_stmt_state, (stp),
      sqlo_stmt_ptr_t  stp 
     )
{
  int status;
  ub4 st = 0;

  /* OCI_ATTR_STMT_STATE is not defined in Oracle < 9i 
   * I define this here, to make it compile with Oracle 8,8i,
   * but nobody should call this 
   * function
   */
#ifndef OCI_ATTR_STMT_STATE
#define OCI_ATTR_STMT_STATE 182
#endif

  if( (status = OCIAttrGet((dvoid*)stp->stmthp, 
                           (ub4)OCI_HTYPE_STMT,
                           (dvoid *)&st, 
                           (ub4 *) 0,
                           (ub4)OCI_ATTR_STMT_STATE, 
                           stp->dbp->errhp)))
    return (status);
  return ((int)st);
}



/*-------------------------------------------------------------------------*/
/**
 * Adds a new database entry to @ref _dbv[].
 * Alllocates a new entry entry if necessary, or reuses an old one.
 * @return The pointer to the entry. The entry is marked used.
 */
static sqlo_db_ptr_t 
DEFUN_VOID(_db_add)
{
  register unsigned int dbv_idx;
  unsigned int first_null_idx;                       /* first null slot */
  bool_t first_null_idx_set;      /* Flag indicates that first_null_idx is valid */
  bool_t found_free;
  unsigned int free_idx;

  TRACE(4, fprintf(_trace_fp, "_db_add starts _dbv_size=%u\n", _dbv_size););

  /* check for initialization */
  if ((! _sqlo_init) || (_dbv_size <= 0))
    return (NULL) ;

  EXEC_WHEN_THREADING(_dbv_lock(););  /* start of critical section */

  /* we need to check first if there is an unused allocated slot */
  found_free = FALSE;
  first_null_idx_set = FALSE;
  free_idx = 0;
  first_null_idx = 0;

  for (dbv_idx = 0; dbv_idx < _dbv_size; ++dbv_idx) {
    if ((_dbv[ dbv_idx ]) && (! _dbv[ dbv_idx ]->used)) {  /* free slot found */
      free_idx = dbv_idx;
      found_free = TRUE;
      break ;
    } else if ( !first_null_idx_set && !_dbv[ dbv_idx ]) {
      first_null_idx     = dbv_idx;           
      first_null_idx_set = TRUE;
    }
  }

  TRACE(4, fprintf(_trace_fp, "_db_add: found_free=%d, free_idx=%u "
                   "first_null_idx_set=%d, first_null_idx=%u\n",
                   found_free, free_idx,
                   first_null_idx_set, first_null_idx););

  /* if an unused allocated slot was found */
  if (found_free) {
    TRACE(3, fprintf(_trace_fp, "_db_add: Reusing db handle %u\n", free_idx););

    _dbv[ free_idx ]->dbh = free_idx;
    _dbv[ free_idx ]->used = TRUE ; /* set the in use flag *before* releasing the mutex */

  } else { 
    /* alloc a new if possible */
    if (first_null_idx_set) {

      TRACE(3, fprintf(_trace_fp,"_db_add: Added new handle %u\n", first_null_idx););
      _dbv[ first_null_idx ] = MALLOC(sizeof(sqlo_db_t)) ;
      
      if (_dbv[ first_null_idx ]) {
        TRACE(4, fprintf(_trace_fp,"_db_add: Allocated %u bytes\n", 
                         (unsigned int) sizeof(sqlo_db_t)););
        memset(_dbv[ first_null_idx ], 0, sizeof(sqlo_db_t)) ;
        
        _dbv[ first_null_idx ]->dbh  =  first_null_idx;
        _dbv[ first_null_idx ]->used = TRUE ; /* set this flag *before* releasing the mutex */
        free_idx = first_null_idx;
      } else {
        /* alloc failed */
        EXEC_WHEN_THREADING(_dbv_unlock(););   /* end of critical section */
        return (NULL);
      }
    } else {
      /* no free slots */
      EXEC_WHEN_THREADING(_dbv_unlock(););   /* end of critical section */
      return (NULL);
    }
  }


  EXEC_WHEN_THREADING(_dbv_unlock();); /* end of critical section */

  return (_dbv[ free_idx ]) ;
}



/*-------------------------------------------------------------------------*/
/**
 * Release a database in _dbv[].
 * Marks it as unused and frees all allocated handles.
 *
 * @param dbp - I The pointer to the database structure
 */
static void
DEFUN(_db_release, (dbp), sqlo_db_ptr_t dbp)
{
  if (!dbp)
    return;

  TRACE(2, fprintf(_get_trace_fp(dbp), "_db_release[%u] starts\n", dbp->dbh); );

  if (dbp->srvhp)
    (void)OCIHandleFree((dvoid *) dbp->srvhp, OCI_HTYPE_SERVER);

  if (dbp->svchp)
    (void)OCIHandleFree((dvoid *) dbp->svchp, OCI_HTYPE_SVCCTX);

  if (dbp->errhp)
    (void)OCIHandleFree((dvoid *) dbp->errhp, OCI_HTYPE_ERROR);

  if (dbp->authp)
    (void)OCIHandleFree((dvoid *) dbp->authp, OCI_HTYPE_SESSION);

  if (dbp->envhp)
    (void)OCIHandleFree((dvoid *) dbp->envhp, OCI_HTYPE_ENV);

  dbp->srvhp = NULL;
  dbp->svchp = NULL;
  dbp->errhp = NULL;
  dbp->authp = NULL;
  dbp->envhp = NULL;

  if (dbp->tnsname) {
    FREE(dbp->tnsname);
    dbp->tnsname = NULL;
  }

  /* close the trace file */
  if (dbp->trace_fp) {
    fclose(dbp->trace_fp);
    dbp->trace_fp = NULL;
  }

  dbp->errcode = 0;
  *(dbp->errmsg) = '\0';
  dbp->status = 0;

  EXEC_WHEN_THREADING( _dbv_lock(); ); /* start of critical section */

  dbp->attached = FALSE;
  dbp->session_created = FALSE;
  dbp->used = FALSE;

  EXEC_WHEN_THREADING( _dbv_unlock(); ); /* end of critical section */
  
}



/*---------------------------------------------------------------------------*/
/**
 * Returns the thread id of this thread.
 * @return The thread id
 */
static inline sqlo_thread_t
DEFUN_VOID(_get_thread_id) 
{
#ifdef ENABLE_PTHREADS
  return pthread_self();
#elif ENABLE_ORATHREADS
  sqlo_thread_t tid = NULL;
  OCIThreadIdInit(_oci_envhp, _oci_errhp, &tid);
  OCIThreadIdGet(_oci_envhp, _oci_errhp, tid);
  return tid;
#endif
  return 0;
}




/*-------------------------------------------------------------------------*/
/**
 * Resets the number of bindpv entries.
 */
static void
DEFUN(_bindpv_reset, (stp), sqlo_stmt_ptr_t  stp)
{
  unsigned int bindp_idx;

  if (stp) {
    for (bindp_idx = 0; bindp_idx < stp->num_bindpv; ++bindp_idx) {
      stp->bindpv[ bindp_idx ] = NULL;
    }

    stp->num_bindpv = 0;

  }

}

/*---------------------------------------------------------------------------*/
/**
 * Compares two thread ids and returns if they are equal
 * @return TRUE if they are equal, FALSE if not.
 */
static inline bool_t
DEFUN(_thread_id_equal, (id1, id2), sqlo_thread_t id1 AND sqlo_thread_t id2)
{
#if ENABLE_ORATHREADS
  boolean result;
  OCIThreadIdSame(_oci_envhp, _oci_errhp, id1, id2, &result);
  return result ? TRUE : FALSE;
#else
  return id1 == id2 ? TRUE : FALSE;
#endif

}



/*-------------------------------------------------------------------------*/
/**
 * Adds a new stmt to @ref _stmtv[].
 * Allocates either a new entry in _stmtv[], or returns an unused.

 * @param dbp I - The database ptr.
 * @return The new entry, marked as used.
 */
static sqlo_stmt_ptr_t 
DEFUN(_get_stmt_ptr, (dbp), const_sqlo_db_ptr_t dbp)
{
  register unsigned stmt_idx;
  unsigned int first_null_idx;                       /* first null slot */
  unsigned int free_idx;
  bool_t found_free;
  bool_t first_null_idx_set;

  register sqlo_stmt_ptr_t *stpp;

  /* check for initialization */
  if ((! _sqlo_init) || (_stmtv_size <= 0))
    return (NULL) ;
  
  EXEC_WHEN_THREADING( _stmtv_lock(); );  /* start of critical section */
  
  /* we need to check first if there is an unused allocated slot */
  first_null_idx_set = FALSE;
  found_free         = FALSE;
  free_idx           = 0;
  first_null_idx     = 0;

  for (stmt_idx = 0, stpp = _stmtv; stmt_idx < _stmtv_size; ++stmt_idx, ++stpp) {
    if ((*stpp) && (! (*stpp)->used)) {  /* free slot found */
      free_idx   = stmt_idx;
      found_free = TRUE;
      break ;
    } else if ( !first_null_idx_set && !*stpp) {
      first_null_idx     = stmt_idx;               /* save the first null slot */
      first_null_idx_set = TRUE;
      break;
    }
  }
  
  /* an unused allocated slot was found */
  if (found_free) {
    stpp = &_stmtv[ free_idx ];
    TRACE(3, fprintf(_get_trace_fp(dbp),
                     "_get_stmt_ptr: Reusing handle %u\n", free_idx););
    (*stpp)->sth = free_idx ;
    (*stpp)->used = TRUE ; /* set this flag *before* releasing the mutex */

  } else { /* we need to find a NULL slot and allocate it */

    if (first_null_idx_set) {  
      TRACE(3, fprintf(_get_trace_fp(dbp), "_get_stmt_ptr: Added new handle %u\n", 
                       first_null_idx););

      stpp = &_stmtv[ first_null_idx ];
      *stpp = MALLOC(sizeof(sqlo_stmt_t)) ;

      if (*stpp) {
        TRACE(4, fprintf(_get_trace_fp(dbp), "_get_stmt_ptr: Allocated %u bytes\n", 
                         (unsigned int)sizeof(sqlo_stmt_t)););
        memset( *stpp, 0, sizeof(sqlo_stmt_t)) ;
        (*stpp)->sth  = first_null_idx ;
        (*stpp)->used = TRUE ; /* set the in use flag *before* releasing the mutex */

      } else {
        EXEC_WHEN_THREADING( _stmtv_unlock(); );   /* end of critical section */
        return (NULL) ;
      }

    } else {
      /* no null slot found */
      EXEC_WHEN_THREADING( _stmtv_unlock(); );   /* end of critical section */
      return (NULL) ;
    }
  }

  EXEC_WHEN_THREADING((*stpp)->thread_id = _get_thread_id(););

  EXEC_WHEN_THREADING(_stmtv_unlock();); /* end of critical section */

  return (*stpp) ;
}


/*-------------------------------------------------------------------------*/
/**
 * Releases a statement.
 * Release is done by setting used to 0.
 */
#ifdef CC_PRAGMA_INLINE
#define PRAGMA INLINE _stmt_release
#endif
static inline void
DEFUN(_stmt_release, (stp), sqlo_stmt_ptr_t  stp)
{
  EXEC_WHEN_THREADING(_stmtv_lock(););  /* enter critical section */

  stp->thread_id = 0;

  stp->used = FALSE;

  EXEC_WHEN_THREADING(_stmtv_unlock();); /* leave critical section */
}


/*-------------------------------------------------------------------------*/
/**
 * Initializes the stmt structure
 * Sets the dbp into stp and copies stmt into the structure.
 * @return SQLO_SUCCESS or SQLO_ERRMALLOC
 */
static int
DEFUN(_stmt_init, (stp, dbp, stmt),
       sqlo_stmt_ptr_t  stp AND
       sqlo_db_ptr_t  dbp AND
       const char *stmt)
{
  unsigned int len = (unsigned int)strlen(stmt) + 1;

  stp->dbp       = dbp;
  stp->stmthp    = NULL;
  stp->stype     = 0;
  stp->opened    = FALSE;
  stp->prepared  = FALSE;
  stp->still_executing = FALSE;
  stp->num_executions  = 0;
  stp->cursor_type = DEFAULT;


  if (!stp->stmt) {
    if ( MIN_STMT_SIZE > len )
      len = MIN_STMT_SIZE;
    
    ERRMALLOC(stp->dbp, stp->stmt, sizeof(char) * len, 
              "_stmt_init(alloc stmt)", SQLO_ERRMALLOC);
    stp->stmt_size = len;
    
  } else if (stp->stmt_size < len ) {
    
    ERREALLOC(stp->dbp, stp->stmt, sizeof(char) * len, 
              "_stmt_init(realloc stmt))", SQLO_ERRMALLOC);
    stp->stmt_size = len;
  }

  strcpy(stp->stmt, stmt);
    
  return SQLO_SUCCESS;
}


/*-------------------------------------------------------------------------*/
/**
 * Creates a new stmt.
 * Allocates the stmt entry via @ref _get_stmt_ptr, allocates the statement handle and
 * initializes the structure.
 * The stmt structure is initialized.
 * 
 * @return SQLO_SUCCESS or < 0 on error.
 */
static int
DEFUN(_stmt_new, (dbp, stmt, stpp),
       sqlo_db_ptr_t  dbp AND
       const char * stmt AND
      sqlo_stmt_ptr_t *stpp)
{
  register sqlo_stmt_ptr_t  stp;
  int status;

  stp = _get_stmt_ptr(dbp);

  if (! stp) {
    sprintf(dbp->errmsg, "SLQORALIB *** FATAL *** : allocation of statement failed") ;
    return SQLO_ERRMALLOC;
  }

  /* init the structure */
  if (SQLO_SUCCESS != (status = _stmt_init(stp, dbp, stmt))) {
    _stmt_release(stp);
    return SQLO_ERROR;
  }

  /* 
   * Allocate the statement handle
   */
  status = OCIHandleAlloc( (dvoid *) dbp->envhp, (dvoid **) &stp->stmthp,
                           OCI_HTYPE_STMT, (size_t) 0, (dvoid **) 0);
  if (status) {
    _stmt_release(stp);
    return status;
  }  
  *stpp = stp;

  return SQLO_SUCCESS;
}


/*-------------------------------------------------------------------------*/
/**
 *
 * (Re)allocates space for stp->bindpv[] and stp->indpv[] if necessary.
 * Sets stp->bindpv_size to the new size.
 * Allocates always in steps of MIN_BINDP, to avoid too many reallocations.
 *
 * @param stp  I - The statement pointer.
 * @param size I - The requested size.
 * 
 * @return <ul>
 * <li>SQLO_SUCCESS on error.
 * <li>SQLO_ERRMALLOC on failure.
 * </ul>
 */
static int 
DEFUN(_alloc_bindp, (stp, size), 
      sqlo_stmt_ptr_t  stp AND 
      unsigned int size)
{
  if (!stp->bindpv_size ) {    /* complety empty ? */

    /* allocate MIN_BINDP to avoid a lot of reallocations */
    if (size < MIN_BINDP)
      size = MIN_BINDP;

    TRACE(4, fprintf(_get_trace_fp(stp->dbp),
                     "_alloc_bindp: alloc sth: %u to %u elements\n",
                     stp->sth, size););

    ERRMALLOC(stp->dbp, stp->bindpv, sizeof(OCIBind*) * size,
              "_alloc_bindp(bindp)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->indpv, sizeof(short) * size, 
              "_alloc_bindp(ind)", SQLO_ERRMALLOC);

    stp->bindpv_size = size;
    _bindpv_reset(stp);

    memset(stp->bindpv, 0, size * sizeof(OCIBind*));
    memset(stp->indpv,  0, size * sizeof(short));

  } else {
    /* Still enough space, or realloc necessary? */
    if (size > stp->bindpv_size) {
      size = 2 * size; /* alloc twice as much, so we don't have to realloc it too often */

      TRACE(4, fprintf(_get_trace_fp(stp->dbp),
                       "_alloc_bindpv: realloc sth: %u to %u elements\n",
                       stp->sth, size););

      ERREALLOC(stp->dbp, stp->bindpv, sizeof(OCIBind*) * size,
                "sqlo_realloc_bind(bindpv)",  SQLO_ERRMALLOC );

      ERREALLOC(stp->dbp, stp->indpv, sizeof(short) * size,  
                "sqlo_realloc_bind(indp)", SQLO_ERRMALLOC );

      stp->bindpv_size = size;
      /* init the new elements */
      size = size - stp->bindpv_size; /* number of elements to be initialized */
      memset(&stp->bindpv[ stp->bindpv_size ], 0, size * sizeof(OCIBind *));
      memset(&stp->indpv[ stp->bindpv_size ],  0, size * sizeof(short));

    }
  }
  return(SQLO_SUCCESS);
}


/*-------------------------------------------------------------------------*/
/**
 * (Re)allocates more space for output variables.
 * (Re)allocates more space for stp->ocolsv[], stp->outv,
 * stp->defnpv, stp->ocol_namev_size and stp->ocol_namev.
 * Sets stp->defnpv_size is set to the new size.
 * Allocates always in steps of MINUM_DEFNPV, to avoid too much memory free/allocs
 *
 * @param stp  I - The statement pointer
 * @param size I - The requested size.
 *
 * @return <ul>
 * <li>SQLO_SUCCESS on error.
 * <li>SQLO_ERRMALLOC on failure.
 * </ul>
 */
static int 
DEFUN(_alloc_definep, (stp, size),  sqlo_stmt_ptr_t  stp AND unsigned int size)
{
  register unsigned int i;

  if (!stp->defnpv_size ) {     /* complety empty ? */

    /* allocate MIN_DEFNP to avoid a lot of reallocations */
    if (size < MIN_DEFNP)
      size = MIN_DEFNP;

    TRACE(4, fprintf(_get_trace_fp(stp->dbp),
                     "_alloc_definep: alloc sth: %u for %u columns\n",
                     stp->sth, size););

    ERRMALLOC(stp->dbp, stp->defnpv, sizeof(OCIDefine *) * size, 
              "_alloc_definep(alloc defnpv)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->ocolsv, sizeof(sqlo_col_t) * size, 
              "_alloc_definep(alloc ocols)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->outv, sizeof(char *) * size, 
              "_alloc_definep(alloc outv)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->outv_size, sizeof(int) * size, 
              "_alloc_definep(alloc outv_size)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->oindv, sizeof(ub2) * size, 
              "_alloc_definep(alloc oindv)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->rlenv, sizeof(ub2) * size, 
              "_alloc_definep(alloc rlenv)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->ocol_namev, sizeof(char *) * size, 
              "_alloc_definep(alloc ocol_namev)", SQLO_ERRMALLOC);

    ERRMALLOC(stp->dbp, stp->ocol_namev_size, sizeof(int) * size, 
              "_alloc_definep(alloc ocol_namev_size)", SQLO_ERRMALLOC);

    memset(stp->defnpv, 0,  sizeof(OCIDefine *) * size);
    memset(stp->ocolsv, 0, sizeof(sqlo_col_t) * size);
    memset(stp->outv, 0, sizeof(char *) * size);
    memset(stp->outv_size, 0, sizeof(int) * size);
    memset(stp->oindv, 0, sizeof(ub2) * size);
    memset(stp->rlenv, 0, sizeof(ub2) * size);
    memset(stp->ocol_namev, 0, sizeof(char *) * size);
    memset(stp->ocol_namev_size, 0, sizeof(int) * size);

    stp->defnpv_size = size;

  } else {
    /* Still enough space, or realloc necessary? */
    if (size > stp->defnpv_size) {

        size = 2 * size; /* alloc twice as much, so we don't have to realloc too often */

      TRACE(4, fprintf(_get_trace_fp(stp->dbp),
                       "_alloc_definep: realloc sth: %u to %u elements\n",
                       stp->sth, size););

      ERREALLOC(stp->dbp, stp->defnpv, sizeof(OCIDefine *) * size, 
                "_alloc_definep(realloc defnpv)", SQLO_ERRMALLOC );

      ERREALLOC(stp->dbp, stp->ocolsv, sizeof(sqlo_col_t) * size,
                "_alloc_definep(realloc ocols)",  SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->outv, sizeof(char *) * size, 
                "_alloc_definep(realloc outv)", SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->outv_size, sizeof(int) * size, 
                "_alloc_definep(realloc outv_size)", SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->oindv, sizeof(ub2) * size, 
                "_alloc_definep(realloc oindv)", SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->rlenv, sizeof(ub2) * size, 
                "_alloc_definep(realloc rlenv)", SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->ocol_namev, sizeof(char *) * size, 
                "_alloc_definep(realloc ocol_namev)", SQLO_ERRMALLOC);

      ERREALLOC(stp->dbp, stp->ocol_namev_size, sizeof(int) * size, 
                "_alloc_definep(realloc ocol_namev_size)", SQLO_ERRMALLOC);

      /* init the new elements */
      i = size - stp->defnpv_size;

      memset(&stp->defnpv[ stp->defnpv_size ], 0,  sizeof(OCIDefine *) * i);

      memset(&stp->ocolsv[ stp->defnpv_size ], 0, sizeof(sqlo_col_t) * i);

      memset(&stp->outv[ stp->defnpv_size ], 0, sizeof(char*) * i);

      memset(&stp->outv_size[ stp->defnpv_size ], 0, sizeof(int) * i);

      memset(&stp->oindv[ stp->defnpv_size ], 0, sizeof(ub2) * i);

      memset(&stp->rlenv[ stp->defnpv_size ], 0, sizeof(ub2) * i);

      memset(&stp->ocol_namev_size[ stp->defnpv_size ], 0, sizeof(int) * i);

      stp->defnpv_size = size;
    }
  }
  return(SQLO_SUCCESS);
}


/*-------------------------------------------------------------------------*/
/**
 * Set our parameters from the environment.
 * @return <ul>
 * <li> SQLO_SUCCESS
 * <li> != 0 on error
 * </ul>
 */
static int
DEFUN_VOID(_sqlo_getenv)
{
  register int i;
  char * ep;
  char vname[MAX_VNAME_LEN+1];

  /* Preset trace file name */
  strcpy(_trace_file, DEFAULT_TRACE_FNAME);
  
  /* 
   * Get the parameters from the environment 
   */
  for ( i = 0; g_params[i].name ; i++) {
    sprintf(vname, "SQLORA_%s", g_params[i].name);
    ep = vname;

    if ( NULL != (ep = getenv(vname)) && strlen (getenv(vname)) ) {
      switch (g_params[i].vtyp) {
      case INTEGER:
        if (g_params[i].value)
          *((int*) g_params[i].value) = atoi (ep); 
        break;
      case STRING:
        if (g_params[i].value)
          strcpy( (char *) g_params[i].value, ep); 
        break;
      default:
        break;
      }
      if (g_params[i].trigger_fct)
        if (SQLO_SUCCESS != g_params[i].trigger_fct(i))
          return SQLO_ERROR;
    }
  }
  return SQLO_SUCCESS;
} 

/*-------------------------------------------------------------------------*/
/**
 * Opens the trace file.
 *
 * If open fails, an error is printed to stderr.
 * If _trace_level is 0 the trace file will be closed.
 * @param dbp I - The databse pointer (NULL allowed for global trace file).
 * @return  Returns SQLO_SUCCESS or SQLO_ERROR.
 */
static int 
DEFUN(_open_trace_file, (dbp), sqlo_db_ptr_t dbp)
{
  /* If dbp is null, we open the global trace file, else we open a session specific trace file */
  if (NULL == dbp) {
    if (_trace_level > 0 && !_trace_fp) {
      if (NULL == (_trace_fp = fopen(_trace_file, "w"))) {
        fprintf(stderr,"Cannot open %s (errno=%d)\n", _trace_file, errno);
        return SQLO_ERROR;
      }
      fprintf(_trace_fp, "\n**** Starting new trace log ****\n");

    } else if ( 0 == _trace_level && _trace_fp ) {
      fclose(_trace_fp); 
      _trace_fp = NULL;
    }
  } else {
    if (_trace_level > 0) {
      char trace_file[MAX_PATH_LEN+1];
      /* Note: we open a session specific trace file, not a dbh specific one, because
       * dbh's may be recycled.
       */
      _session_count++;              /* increase the session counter */
      /* maybe the dbp holds already a open fp. Close it first */
      if (dbp->trace_fp)
        fclose(dbp->trace_fp);

      sprintf(trace_file, "%s%u", _trace_file, _session_count);
      
      if (NULL == (dbp->trace_fp = fopen(trace_file, "w"))) {
        fprintf(stderr,"Cannot open %s (errno=%d)\n", trace_file, errno);
        return SQLO_ERROR;
      }
      fprintf(dbp->trace_fp, 
              "\n**** Starting new trace log for dbh=%u session=%u ****\n", 
              dbp->dbh, _session_count);
    }
  }
  return SQLO_SUCCESS;
}


/*-------------------------------------------------------------------------*/
/**
 * Saves the error message for dbp->status.
 *
 * This function is called by @ref CHECK_OCI_STATUS_RETURN if dbp->status != 0.
 * If dbp->status is OCI_ERROR, dbp->errcode is set to the Oracle error code.
 * @param dbp    I - The database pointer
 * @param action I - A string identifying the action that lead to the status
 * @param object I - A string identifying the object where the action was executed on.
 * @param lineno I - The line in the source code, where the error occured.
 *
 * @return SQLO_SUCCESS
 */
static int
DEFUN(_save_oci_status, (dbp, action, object, lineno),
      sqlo_db_ptr_t  dbp AND 
      const char *action AND
      const char *object
      AND int lineno)
{
  char errbuf[SQLO_MAX_ERRMSG_LEN+1];
  unsigned int len;

  if (!dbp)
    return 0;

  if (!object)
    object = "";

  *errbuf = '\0';
  dbp->errcode = dbp->status;   /* preset with something usefull */
  TRACE(3, 
        if (dbp->status != OCI_SUCCESS) {
          fprintf(_get_trace_fp(dbp),"_save_oci_status: %d\n", dbp->status);
        }
        );
  switch (dbp->status) {

  case OCI_SUCCESS:
    break;
  
  case OCI_SUCCESS_WITH_INFO:
    (void) OCIErrorGet(dbp->errhp, (ub4) 1, (text *) NULL,
                       &dbp->errcode,
                       (text *)errbuf, (ub4) sizeof(errbuf), 
                       OCI_HTYPE_ERROR);
    sprintf(dbp->errmsg, 
            "%s\n(sqlora line: %d)\n", errbuf, lineno);
    break;

  case OCI_NEED_DATA:
    sprintf(dbp->errmsg, 
            "ERROR: OCI_NEED_DATA\n(sqlora line: %d)\n", lineno);
    break;

  case OCI_NO_DATA:
    sprintf(dbp->errmsg, 
            "ERROR: OCI_NO_DATA\n(sqlora line: %d)\n", lineno);
    break;

  case OCI_ERROR:

    (void)OCIErrorGet(dbp->errhp, (ub4) 1, (text *) NULL, &dbp->errcode,
                       (text *)errbuf, (ub4) sizeof(errbuf), OCI_HTYPE_ERROR);
    sprintf(dbp->errmsg, 
            "%s\n(sqlora line: %d)\n", errbuf, lineno);

    break;

  case OCI_INVALID_HANDLE:
    sprintf(dbp->errmsg, 
            "ERROR: OCI_INVALID_HANDLE\n(sqlora line: %d)\n", lineno);
    break;

  case OCI_STILL_EXECUTING:
    sprintf(dbp->errmsg, 
            "ERROR: OCI_STILL_EXECUTING\n(sqlora line: %d)\n", lineno);
    break;

  case OCI_CONTINUE:
    sprintf(dbp->errmsg, 
            "ERROR: OCI_CONTINUE\n(sqlora line: %d)\n", lineno);
    break;

  case SQLO_INVALID_DB_HANDLE:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Invalid database handle.\n(sqlora line: %d)\n", 
            dbp->status, lineno);
    break;
  case SQLO_INVALID_STMT_HANDLE:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Invalid statement handle.\n(sqlora line: %d)\n", 
            dbp->status, lineno);
    break;
  case SQLO_STMT_NOT_OPENED:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Cursor is not open.\n(sqlora line: %d)\n", 
            dbp->status, lineno);
    break;
  case SQLO_STMT_NOT_PARSED:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Stmt is not prepared.\n(sqlora line: %d)\n", 
            dbp->status, lineno);
    break;
  case SQLO_INVALID_STMT_TYPE:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Sorry, this function cannot handle your type of statement.\n"
            "(sqlora line: %d)\n", dbp->status, lineno);
    break;

  case SQLO_INVALID_SQL:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8%05d: Invalid SQL parsed.\n"
            "(sqlora line: %d)\n", dbp->status, lineno);
    break;
  case SQLO_ERRMALLOC:
    /* concatenate the error message. */
    sprintf(&dbp->errmsg[strlen(dbp->errmsg)], 
            "LIBSQLORA8%05d: Memory allocation error.\n(sqlora line: %d)\n", dbp->status,
            lineno);
    break;

  default:
    sprintf(dbp->errmsg, 
            "LIBSQLORA8-00000: Unknown status %d\n(sqlora line: %d)\n", dbp->status,
            lineno);
    break;
  }

  if ((len = strlen (dbp->errmsg) + strlen (action) + strlen (object) + 40)
      > SQLO_MAX_ERRMSG_LEN)
    {
      len = SQLO_MAX_ERRMSG_LEN - strlen (dbp->errmsg) - strlen (action) - 40;
    }
  else
    len = (unsigned int) strlen(object);

  if (strlen(action) ) {
    sprintf (&dbp->errmsg[strlen(dbp->errmsg)], "\nSQL error while doing %s", action);

    if (len) {
      sprintf (&dbp->errmsg[strlen(dbp->errmsg)], " on:\n\"%*.*s\"", (int) len, 
               (int) len, object);
    }
  }
  strcat(dbp->errmsg,"\n");
  TRACE(1, (void) fputs(dbp->errmsg, _get_trace_fp(dbp)););
  strcpy(_errmsg, dbp->errmsg);
  return (SQLO_SUCCESS);
}


/*---------------------------------------------------------------------------*/
/**
 * Returns the statement type as a string.
 * @param stype The statement type
 * @return A constant string telling you the kind of statement.
 */
static const char *
DEFUN(_get_stmt_type_str, (stype), int stype)
{
  switch (stype) {
  case OCI_STMT_SELECT: return "SELECT"; break;
  case OCI_STMT_UPDATE: return "UPDATE"; break;
  case OCI_STMT_DELETE: return "DELETE"; break;
  case OCI_STMT_INSERT: return "INSERT"; break;
  case OCI_STMT_CREATE: return "CREATE"; break;
  case OCI_STMT_DROP: return "DROP"; break;
  case OCI_STMT_ALTER: return "ALTER"; break;
  case OCI_STMT_BEGIN: return "BEGIN"; break;
  case OCI_STMT_DECLARE: return "DECLARE"; break;
  default:
    return "UNKNOWN";
    break;
  }
}


/*---------------------------------------------------------------------------*/
/**
 * Returns the data type as string
 * @param dtype The data type
 * @return A constant string telling you the data type.
 */
static const char *
DEFUN(_get_data_type_str, (dtype), int dtype)
{
  /* the constants are defined in ocidfn.h */
  switch (dtype) {
  case SQLT_CHR: return "character string"; break;
  case SQLT_NUM: return "oracle numeric"; break;
  case SQLT_INT: return "integer"; break;
  case SQLT_FLT: return "floating point number"; break;
  case SQLT_STR: return "zero terminated string"; break;
  case SQLT_VNU: return "num with preceding length byte"; break;
  case SQLT_PDN: return "packed decimal numeric"; break;
  case SQLT_LNG: return "long"; break;
  case SQLT_VCS: return "variable character string"; break;
  case SQLT_NON: return "Null/empty PCC Descriptor entry "; break;
  case SQLT_RID: return "rowid"; break;
  case SQLT_DAT: return "date in oracle format"; break;
  case SQLT_VBI: return "binary in VCS format"; break;
  case SQLT_BIN: return "binary data(DTYBIN)"; break;
  case SQLT_LBI: return "long binary"; break;
  case SQLT_UIN: return "unsigned integer"; break;
  case SQLT_SLS: return "dispay sign leading separate"; break;
  case SQLT_LVC: return "longer longs (char)"; break;
  case SQLT_LVB: return "longer longs (binary)"; break;
  case SQLT_AFC: return "ansi fixed char"; break;
  case SQLT_AVC: return "ansi var char"; break;
  case SQLT_CUR: return "cursor type"; break;
  case SQLT_RDD: return "rowid descriptor"; break;
  case SQLT_LAB: return "label type"; break;
  case SQLT_OSL: return "oslabel type"; break;
  case SQLT_NTY: return "named object type"; break;
  case SQLT_REF: return "ref type"; break;
  case SQLT_CLOB: return "character lob"; break;
  case SQLT_BLOB: return "binary lob"; break;
  case SQLT_BFILEE: return "binary file lob"; break;
  case SQLT_CFILEE: return "character file lob"; break;
  case SQLT_RSET: return "result set type"; break;
  case SQLT_NCO: return "named collection type"; break;
  case SQLT_VST: return "OCIString type"; break;
  case SQLT_ODT: return "OCIDate type"; break;
  default:
    return "UNKNOWN";
    break;
  }
}


/*---------------------------------------------------------------------------*/
/** 
 * Strips all trailing blanks in s.
 * @param s   I - The string to strip
 * @param len I - The current string length.
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _strip_string
#endif
static inline void 
DEFUN(_strip_string, (s, len), char * s AND unsigned int len )
{
  register char *p;

  for ( p = &s[len - 1]; len > 0 && ' ' == *p ; --len)
     *p = '\0';
}

/*---------------------------------------------------------------------------*/
/**
 * Returns the last error code. 
 * Saves it also in dbp->errcode 
 *
 * @param dbp I - The database pointer.
 * @return The Oracle error code.
 */
#if CC_PRAGMA_INLINE
#pragma INLINE _get_errcode
#endif
static inline int 
DEFUN(_get_errcode, (dbp), sqlo_db_ptr_t  dbp)
{
  assert( (dbp != NULL) );

  (void)OCIErrorGet(dbp->errhp, 
                    (ub4) 1,
                    (text *) NULL,
                    &dbp->errcode,
                    NULL, 
                    (ub4) 0, 
                    OCI_HTYPE_ERROR);
  
  return dbp->errcode;
}


/*---------------------------------------------------------------------------*/
/** 
 * Binds the input variables.
 * This is the same as @ref sqlo_bind_by_pos, except that the stmt pointer is passed
 * instead of the stmt handle.
 *
 * @return SQLO_SUCCESS or SQLO_ERROR.
 * @see sqlo_bind_by_pos
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _bind_by_pos
#endif
static inline int 
DEFUN(_bind_by_pos, 
       (stp, param_pos, param_type, param_addr, param_size, ind_addr, is_array),
      sqlo_stmt_ptr_t  stp AND 
      unsigned int param_pos AND 
      int param_type AND 
      const void * param_addr AND 
      unsigned int  param_size AND 
      short * ind_addr 
      AND int is_array)
{
  register int status;
  register OCIBind ** bindp_addr;

  TRACE(3, 
        fprintf(_get_trace_fp(stp->dbp), "_bind_by_pos pos: %u type: %d (%s), size: %u\n",
                param_pos, param_type, _get_data_type_str(param_type),
                param_size););
  
  if (stp->prepared) {
    /* make sure we have enough memory */
    if ( param_pos > stp->bindpv_size) {
      status = _alloc_bindp(stp, param_pos);
        
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos. alloc error", "");
    }

    bindp_addr = &stp->bindpv[ param_pos - 1 ];

    status = OCIBindByPos(stp->stmthp,
                          bindp_addr,
                          stp->dbp->errhp,
                          (ub4) param_pos,
                          (ub1 *) param_addr,
                          (sword) param_size,
                          param_type,
                          (dvoid *) ind_addr,
                          (ub2 *) 0,
                          (ub2) 0,
                          (ub4) 0,
                          (ub4 *) 0,
                          OCI_DEFAULT);
      
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos. Cannot bind", NULL);
    if (is_array) {
      status = OCIBindArrayOfStruct(*bindp_addr,
                                    stp->dbp->errhp, 
                                    param_size,
                                    ind_addr ? sizeof(short) : 0,
                                    0,
                                    0);
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos. BindArrayOfStruct", "");
    }

    if (param_pos > stp->num_bindpv)
      stp->num_bindpv = param_pos;

  } else {
    status = SQLO_STMT_NOT_PARSED;
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------*/
/** 
 * Bind input variables.
 * This is the same as @ref sqlo_bind_by_pos2, except that the statement pointer
 * is passed here instead of the statement handle.
 * @return <ul>
 * <li> SQLO_SUCCESS
 * <li> < 0 on error.
 * </ul>
 * @see sqlo_bind_by_pos2
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _bind_by_pos2
#endif
static inline int 
DEFUN(_bind_by_pos2, 
      (stp, param_pos, param_type, param_addr, param_size, ind_addr, rcode_addr, skip_size),
      sqlo_stmt_ptr_t  stp AND
      unsigned param_pos AND 
      int param_type AND 
      const void * param_addr AND 
      unsigned int  param_size AND 
      short * ind_addr AND 
      unsigned short * rcode_addr AND
      unsigned int skip_size)
{
  register int status;
  register OCIBind ** bindp_addr;

  TRACE(3, 
        fprintf(_get_trace_fp(stp->dbp), 
                "_bind_by_pos pos2: %u type: %d (%s), size: %u\n",
                param_pos, param_type, _get_data_type_str(param_type),
                param_size););
  
  if (stp->prepared) {
    /* make sure we have enough memory */
    if ( param_pos > stp->bindpv_size) {
      status = _alloc_bindp(stp, param_pos);
        
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos2. alloc error", "");
    }

    bindp_addr = &stp->bindpv[ param_pos - 1 ];

    status = OCIBindByPos(stp->stmthp,
                          bindp_addr,
                          stp->dbp->errhp,
                          (ub4) param_pos,
                          (ub1 *) param_addr,
                          (sword) param_size,
                          (ub2) param_type,
                          (dvoid *) ind_addr,
                          (ub2 *) 0,
                          (ub2 *) rcode_addr,
                          (ub4) 0,
                          (ub4 *) 0,
                          OCI_DEFAULT);
      
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos. Cannot bind", NULL);
    if (skip_size) {
      status = OCIBindArrayOfStruct(*bindp_addr,
                                    stp->dbp->errhp, 
                                    skip_size,
                                    ind_addr ? skip_size : 0,
                                    0,
                                    rcode_addr ? skip_size : 0);

      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_by_pos2. BindArrayOfStruct", "");
    }

    if (param_pos > stp->num_bindpv)
      stp->num_bindpv = param_pos;

  } else {
    status = SQLO_STMT_NOT_PARSED;
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------*/
/** 
 * Defines an output variable.
 * The same as @ref sqlo_define_by_pos2, except that the stmt pointer is passed
 * here instead of the statement handle.
 * @see sqlo_define_by_pos2
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _define_by_pos2
#endif
static inline int 
DEFUN(_define_by_pos2, 
       (stp, value_pos, value_type, value_addr, value_size, ind_addr, 
        rlen_addr, rcode_addr, skip_size),
       sqlo_stmt_ptr_t  stp AND
       unsigned int value_pos AND 
       int value_type AND 
       const void * value_addr 
       AND unsigned int  value_size AND 
       short * ind_addr AND
       ub2 * rlen_addr AND /* where the returned length is stored. */
       ub2 * rcode_addr AND /* usefully in array fetch. contains the status */
       unsigned int skip_size)
{
  int status;

  TRACE(3, 
        fprintf(_get_trace_fp(stp->dbp), 
                 "_define_by_pos2 pos: %u type: %d (%s), size: %u, skip_size: %u\n",
                 value_pos, value_type, _get_data_type_str(value_type),
                 value_size, skip_size););
  
  if (stp->prepared) {

    status = _alloc_definep(stp, value_pos);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos2. alloc error", "");

    /* save the datatype */
    stp->ocolsv[value_pos -1].dtype = (ub2)value_type;

    status = OCIDefineByPos(stp->stmthp,
                            &stp->defnpv[ value_pos - 1 ],
                            stp->dbp->errhp,
                            (ub4) value_pos,
                            (ub1 *) value_addr,
                            (sword) value_size,
                            value_type,
                            (dvoid *) ind_addr,
                            (ub2 *) rlen_addr,
                            (ub2 *) rcode_addr,
                            OCI_DEFAULT);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos2: Cannot define", NULL);
    
    if (skip_size) {
      status = OCIDefineArrayOfStruct(stp->defnpv[ value_pos - 1 ],
                                      stp->dbp->errhp, 
                                      skip_size,
                                      ind_addr ? skip_size : 0,
                                      rlen_addr ? skip_size : 0,
                                      rcode_addr ? skip_size : 0
                                      );
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos2 BindArrayOfStruct", "");
    }

    stp->num_defnpv++;

  } else {
    status = SQLO_STMT_NOT_PARSED;
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------*/
/** 
 * Defines an output variable.
 * The same as @ref sqlo_define_by_pos, except that the stmt pointer is passed
 * here instead of the statement handle.
 * @see _define_by_pos
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _define_by_pos
#endif
static inline int 
DEFUN(_define_by_pos, 
       (stp, value_pos, value_type, value_addr, value_size, ind_addr, 
        rlen_addr, rcode_addr, is_array),
       sqlo_stmt_ptr_t  stp AND
       unsigned int value_pos AND 
       int value_type AND 
       const void * value_addr 
       AND unsigned int  value_size AND 
       short * ind_addr AND
       ub2 * rlen_addr AND /* where the returned length is stored. */
       ub2 * rcode_addr AND /* usefully in array fetch. contains the status */
       int is_array)
{
  int status;
  TRACE(3, 
        fprintf(_get_trace_fp(stp->dbp), 
                "_define_by_pos pos: %u type: %d (%s), size: %u, isa: %d\n",
                value_pos, value_type, _get_data_type_str(value_type),
                value_size, is_array););

  if (stp->prepared) {

    status = _alloc_definep(stp, (unsigned int)value_pos);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos. alloc error", "");

    /* save the datatype */
    stp->ocolsv[value_pos -1].dtype = (ub2)value_type;

    status = OCIDefineByPos(stp->stmthp,
                            &stp->defnpv[ value_pos - 1 ],
                            stp->dbp->errhp,
                            (ub4) value_pos,
                            (ub1 *) value_addr,
                            (sword) value_size,
                            value_type,
                            (dvoid *) ind_addr,
                            (ub2 *) rlen_addr,
                            (ub2 *) rcode_addr,
                            OCI_DEFAULT);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos: Cannot define", NULL);
    
    if (is_array) {
      status = OCIDefineArrayOfStruct(stp->defnpv[ value_pos - 1 ],
                                      stp->dbp->errhp, 
                                      value_size,
                                      ind_addr ? sizeof(short) : 0,
                                      rlen_addr ? sizeof(short) : 0,
                                      rcode_addr ? sizeof(short) : 0
                                      );
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_define_by_pos BindArrayOfStruct", "");
    }

    stp->num_defnpv++;

  } else {
    status = SQLO_STMT_NOT_PARSED;
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);

}


/*---------------------------------------------------------------------------*/
/**
 * Binds the input variables for sqlo_open, sqlo_open2
 * Treats an empty string and a NULL pointer in argv as NULL.
 * Call @ref _bind_by_pos for each entry in argv.
 *
 * @param stp  I - Statement pointer
 * @param argc I - Number of entries in argv
 * @param argv I - bind values
 *
 * @return SQLO_SUCCESS or < 0 on error.
 */
static int 
DEFUN(_bind_argv, (stp, argc, argv),
       sqlo_stmt_ptr_t  stp AND
       unsigned int argc AND
       const char ** argv)
{
  register unsigned int arg_idx;
  int status = SQLO_SUCCESS;
  register const char ** arg;
  unsigned int size;
  register short * ind_ptr;


  /* allocate the necessary intput bind variables */
  if (argc > stp->bindpv_size) {
    status = _alloc_bindp(stp, argc);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "Malloc error in _bind_argv", "");
  }

  arg = argv;
  ind_ptr = stp->indpv;
  for (arg_idx = 0 ; arg_idx < argc; ++arg_idx, ++arg, ++ind_ptr) {
    if (!*arg || !**arg ) {    /* treat null pointer or empty string as null */
      *ind_ptr = SQLO_NULL_IND;
      size = 0;
    } else {
      *ind_ptr = SQLO_NOT_NULL_IND;
      size = strlen(*arg) + 1;
    }

    status = _bind_by_pos(stp, arg_idx + 1, SQLOT_STR, *arg, 
                          size, ind_ptr, 0);

    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_bind_argv. Cannot bind:",
               *arg);
        
  }
  return (status);
}


/*---------------------------------------------------------------------------*/
/**
 * Calculates the size of the output buffer for a datatype.
 *
 * @param bufsizep  O - The calculated size.
 * @param data_type I - The oracle data type
 * @param prec      I - The precision (number type)
 * @param scale     I - The scale (number type)
 * @param dbsize    I - Size in the database
 * 
 * @return SQLO_SUCCESS or SQLO_ERROR for a datatype that could not be
 *         handled here (LOB).
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _calc_obuf_size
#endif
static inline int 
DEFUN(_calc_obuf_size, (bufsizep, data_type, prec, scale, dbsize), 
      unsigned int * bufsizep AND
      unsigned int data_type AND
      int prec AND
      int scale AND
      unsigned int dbsize
      )
{

  unsigned int buffer_size = 0;
  int status = SQLO_SUCCESS;

  switch(data_type) {

  case SQLT_NUM:
  case SQLT_INT:
  case SQLT_FLT:

    if (scale > prec) {
      buffer_size = (unsigned int) scale + 3; /* sign, comma and \0 */
    } else if (prec > 0) {
      buffer_size = (unsigned int) prec + 3; 
    } else {
      buffer_size = (unsigned int) (2 * dbsize) + 3;
    }

    /* use a minimum buffer */
    if (buffer_size < (2 * dbsize) + 3 )
      buffer_size = ( 2* dbsize)  + 3;

    break;

  case SQLT_CHR:
  case SQLT_STR:
  case SQLT_AFC:
    buffer_size = (dbsize + 1);
    break;
  case SQLT_RID:
  case SQLT_RDD:
    buffer_size = 32;
    break;
  case SQLT_DAT:
    buffer_size=  40; /* just big */
    break;

  case SQLT_LNG:
    buffer_size = _max_long_size;

    break;

  case SQLT_BLOB:
  case SQLT_CLOB:
  case SQLT_BFILEE:
  case SQLT_CFILEE:

    status = SQLO_ERROR;        /* not supported in this mode */
    break;

  default:
    buffer_size = (( 2 * dbsize) + 1);
  }

  *bufsizep = buffer_size;

  return status;

}


/*---------------------------------------------------------------------------*/
/** 
 * Defines one output column
 *
 * Describe one output variable and create the buffer to store the query result
 * @param stp  I - The statement handle pointer
 * @param colp I - Pointer to the column structure
 * @param pos  I - The column position (1 based).
 *
 * @return SQLO_SUCCESS or < 0 on error.
 */
#ifdef CC_PRAGMA_INLINE
#pragma INLINE _define_ocol_by_pos
#endif
static inline int 
DEFUN(_define_ocol_by_pos, (stp, colp, pos), 
      sqlo_stmt_ptr_t stp AND 
      sqlo_col_t * colp AND
      unsigned int pos)
{
  int status = OCI_SUCCESS;
  sqlo_db_ptr_t  dbp = stp->dbp;
  char * col_name;
  unsigned int col_name_len;
  unsigned int buffer_size;
  unsigned int cur_col_name_len;
  register unsigned int col_idx = pos - 1;     /* the index into our arrays */
  ub2 data_type;                /* the database datatype */

  TRACE(3, fprintf(_get_trace_fp(dbp), "Define output column number %u\n", pos););

  /* save index and link to the statement */
  colp->pos = col_idx;
  colp->stp = (struct _sqlo_stmt_t *)stp;

  TRACE(3, fprintf(_get_trace_fp(dbp), "Getting parameters of col: %u\n", pos););

  status = OCIParamGet(stp->stmthp, 
                       OCI_HTYPE_STMT, 
                       dbp->errhp,
                       (dvoid *)&colp->paramd, 
                       (ub4) pos);

  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_ocol_by_pos", "OCIParamGet");

  /*--------------*/
  /* get datatype */
  /*--------------*/
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &data_type,(ub4 *) 0, 
                      (ub4) OCI_ATTR_DATA_TYPE, 
                      (OCIError *)dbp->errhp );
  
  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_by_pos", "OCIAttrGet(datatype)");
  TRACE(3, fprintf(_get_trace_fp(dbp), "datatype: %d (%s)\n", (int) data_type,
                   _get_data_type_str((int)data_type)););


  /*-----------------*/
  /* get column name */
  /*-----------------*/
  col_name_len = 0;
  col_name = NULL;
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid**) &(col_name),
                      (ub4 *) &(col_name_len), 
                      (ub4) OCI_ATTR_NAME, 
                      (OCIError *) dbp->errhp);

  cur_col_name_len = col_name_len; /* save the real length */
  CHECK_OCI_STATUS_RETURN(dbp, status, 
             "_define_by_pos", "OCIAttrGet(column_name)");

  if (!colp->col_name) {

    if (col_name_len < MIN_COL_NAME_LEN)
      col_name_len = MIN_COL_NAME_LEN;

    ERRMALLOC(stp->dbp, colp->col_name, 
              sizeof(char) * (col_name_len + 1),
              "_define_by_pos(alloc col_name)", SQLO_ERRMALLOC);
    colp->col_name_size = col_name_len;

  } else {

    if (col_name_len > colp->col_name_size) {
      col_name_len = 2 * col_name_len; /* alloc always twice as much we need */
      ERREALLOC(stp->dbp, colp->col_name, 
                sizeof(char) * (col_name_len + 1),
                "_define_by_pos(realloc col_name)", SQLO_ERRMALLOC);

      colp->col_name_size = col_name_len;
    }
  }

  strncpy(colp->col_name, col_name, (size_t)cur_col_name_len);
  colp->col_name[cur_col_name_len] = '\0';


  TRACE(3, fprintf(_get_trace_fp(dbp), "colname: %.*s (colname_len: %u)\n", 
                   (int)colp->col_name_size, colp->col_name, cur_col_name_len););


  /* Point our ocol_namev[i] to the right column name.
   * The name can be fetched by sqlo_ocol_names();
   * Store the length of the column name in ocol_namev_size[i]*/
  stp->ocol_namev[ col_idx ] = colp->col_name;
  stp->ocol_namev_size[ col_idx ] = cur_col_name_len;


  /*-------------*/
  /* get db size */
  /*-------------*/
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &(colp->dbsize),
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_DATA_SIZE, 
                      (OCIError *) dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, 
             "_define_by_pos", "OCIAttrGet(dbsize)");

  TRACE(3, fprintf(_get_trace_fp(dbp), "dbsize: %d\n", (int) colp->dbsize););


  /*---------------*/
  /* get precision */
  /*---------------*/
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &(colp->prec),
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_PRECISION, 
                      (OCIError *) dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, 
             "_define_by_pos", "OCIAttrGet(precision)");

  TRACE(3, fprintf(_get_trace_fp(dbp), "prec: %d\n", (int) colp->prec););


  /*-----------*/
  /* get scale */
  /*-----------*/
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &(colp->scale),
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_SCALE, 
                      (OCIError *) dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_by_pos", "OCIAttrGet(scale)");

  TRACE(3, fprintf(_get_trace_fp(dbp), "scale: %d\n", (int) colp->scale););

  /*-------------*/
  /* get null ok */
  /*-------------*/
  status = OCIAttrGet((dvoid*) colp->paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &(colp->nullok),
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_IS_NULL, 
                      (OCIError *) dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, 
             "_define_by_pos", "OCIAttrGet(is null)");

  TRACE(3, fprintf(_get_trace_fp(dbp), "nullok: %d\n", (int) colp->nullok););
      
  /* calculate the required output buffer size */
  status = _calc_obuf_size(&buffer_size, data_type, colp->prec, colp->scale, colp->dbsize);
  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_by_pos", "_calc_obuf_size");

  TRACE(3, fprintf(_get_trace_fp(dbp), "buffer size: %u\n", buffer_size););

  if (!stp->outv[ col_idx ]) {
    if (buffer_size) {
      if (buffer_size < MIN_OBUF_SIZE)
        buffer_size = MIN_OBUF_SIZE;
      ERRMALLOC(dbp, stp->outv[ col_idx ],(buffer_size)* (unsigned int)sizeof(char), 
                "_define_by_pos(alloc outv[i])", SQLO_ERRMALLOC);
    }
    stp->outv_size[ col_idx ] = buffer_size;
  } else {
    if (buffer_size > stp->outv_size[ col_idx ]) {
      if (buffer_size < _max_long_size) /* don't double for long columns */
        buffer_size = 2 * buffer_size; /* alloc always twice as much to avoid lots of reallocs */
      ERREALLOC(dbp, stp->outv[ col_idx ], (buffer_size)* (unsigned int)sizeof(char), 
                "_define_by_pos(realloc outv[col_idx])", SQLO_ERRMALLOC);
      stp->outv_size[ col_idx ] = buffer_size;
    }
  }

  /*--------------------*/
  /* Define this column */
  /*--------------------*/
  status = _define_by_pos(stp, 
                           pos, 
                           SQLT_STR, 
                           stp->outv[ col_idx ],
                           stp->outv_size[ col_idx ],
                           (short *) &stp->oindv[ col_idx ],
                           &stp->rlenv[ col_idx ],
                           NULL,
                           0);
  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_by_pos", "_define_by_pos");

  /* free the descriptor */
  status = OCIDescriptorFree(colp->paramd, OCI_DTYPE_PARAM);
  colp->paramd = NULL;
  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_by_pos", "OCIDescriptorFree(paramd)");

  return (status);
}


/*---------------------------------------------------------------------------*/
/** 
 * Define the output variables.
 *
 * Describe the output variables and create the buffers for the ouptut.
 *
 * @return SQLO_SUCCESS or < 0 on error.
 */
static int 
DEFUN(_define_output, (stp), sqlo_stmt_ptr_t  stp)
{
  register unsigned int col_pos; /* The column position (1based) */
  int status;
  ub4 num_cols;                 /* number of columns in the select list */
  sqlo_col_ptr_t colp;
  sqlo_db_ptr_t  dbp = stp->dbp;
  

  /* Already defined ? */
  if (stp->num_defnpv)
    return (SQLO_SUCCESS);

  /* Describe the output variables.
   * REFCURSORs are already exectuted by there parent stmt
   */
  
  if ( 0 == stp->num_executions && !(stp->cursor_type == REFCURSOR) ) {
    if (OCI_STILL_EXECUTING == 
        (status = OCIStmtExecute(dbp->svchp, 
                                 stp->stmthp, 
                                 dbp->errhp, 
                                 (ub4) 0, 
                                 (ub4) 0, 
                                 (OCISnapshot *) 0, 
                                 (OCISnapshot *) 0,
                                 (ub4) OCI_DEFAULT)
         )
        )
      {
        return status;
      } else {
        CHECK_OCI_STATUS_RETURN(dbp, status, "_define_output", "OCIStmtExecute(DESCRIBE)");
      }
    stp->num_executions++;
  }

  if (stp->cursor_type == REFCURSOR)
    stp->opened = TRUE;

  /* Get info about the select list */
  status = OCIAttrGet((dvoid*)stp->stmthp,
                      (ub4) OCI_HTYPE_STMT, 
                      (dvoid *)  &num_cols, 
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_PARAM_COUNT, 
                      dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_output", "OCIAttrGet(PARAM_COUNT)");
  TRACE(3, fprintf(_get_trace_fp(dbp), "Number of columns in select list: %d\n", (int) num_cols););
  
                                /* allocate the space for the output */
  status = _alloc_definep(stp, num_cols);
  CHECK_OCI_STATUS_RETURN(dbp, status, "_define_output", "Memory allocation error");

  /* define all columns */
  for (col_pos = 1, colp = stp->ocolsv; col_pos <= (unsigned int) num_cols ; 
       ++col_pos, ++colp) {

    if (SQLO_SUCCESS != (status = _define_ocol_by_pos(stp, colp, col_pos)))
      return (status);

  }

  return (status);
}


/*-------------------------------------------------------------------------*/
/**
 * Close all open cursors which are in the still executing status
 * This function returns no status code and all errors returned by @ref sqlo_close
 * are ignored.
 * @param dbp  I - The database pointer
 */
static void 
DEFUN(_close_all_executing_cursors, (dbp), const_sqlo_db_ptr_t dbp)
{
  unsigned int stmt_idx;

  for (stmt_idx = 0; stmt_idx < _stmtv_size; ++stmt_idx) {

    if (_stmtv[ stmt_idx ] != NULL) {

      sqlo_stmt_ptr_t stp = _stmtv[ stmt_idx ];

      if (stp->dbp == dbp && stp->used && stp->still_executing ) {
        if ( THREADS_ENABLED && OCI_THREADED == _oci_init_mode ) {
          if ( TRUE == _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
            sqlo_close((int) stp->sth);
          }
        } else {
            sqlo_close((int) stp->sth);
        }
      } /* endif is valid and executing */

    } /* endif is stmt_idx pointing to a allocated entry */

  } /* end for stmt_idx */
}

/**
 * Close all open cursors on a connection.
 * This function returns no status code and all errors returned by @ref sqlo_close
 * are ignored.
 * @param dbp  I - The database pointer
 */
static void 
DEFUN(_close_all_db_cursors, (dbp), const_sqlo_db_ptr_t dbp)
{
  unsigned int stmt_idx;

  /*  EXEC_WHEN_THREADING( _stmtv_lock(); );*/
  for (stmt_idx = 0; stmt_idx < _stmtv_size; ++stmt_idx) {
    if (_stmtv[ stmt_idx ] != NULL) {

      sqlo_stmt_ptr_t stp = _stmtv[ stmt_idx ];

      if (stp->dbp == dbp && stp->used ) {
        /* make sure we close only cursors in our thread */
        if ( THREADS_ENABLED && OCI_THREADED == _oci_init_mode) {
          if ( TRUE == _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
            sqlo_close((int) stp->sth);
          }
        } else {
            sqlo_close((int) stp->sth);
        } /* end if thread ids are equal */
      } /* end if the same database and in use */
    } /* entry not null */
  } /* end for */
}


/*---------------------------------------------------------------------------*/
/**	
 * Gets the blocking mode for a dbp.
 * @param dbp           I - The database pointer.
 * @param blocking      O - SQLO_ON if in blocking mode, SQLO_OFF if in non-blocking 
 *                          mode
 *
 * @return <ul>
 * <li>SQLO_SUCCESS 
 * <li>SQLO_ERROR
 * </ul>
 */
static int 
DEFUN(_get_blocking_mode, (dbp, blocking), 
      sqlo_db_ptr_t dbp AND  unsigned int * blocking)
{
  int status;
  ub1 non_blocking = 0;

  /* Returns TRUE if the server is in non-blocking mode */
  status = OCIAttrGet((dvoid*)dbp->srvhp,
                      (ub4) OCI_HTYPE_SERVER, 
                      (dvoid *)  &non_blocking, 
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_NONBLOCKING_MODE, 
                      dbp->errhp);

  TRACE(3, fprintf(_get_trace_fp(dbp), "_get_blocking_mode: status=%d, non_blocking=%u\n", 
                   status, non_blocking););

  if (OCI_SUCCESS == status )
    {
      if (non_blocking)
        *blocking = SQLO_OFF;
      else
        *blocking = SQLO_ON;
    }
  else
    status = SQLO_ERROR;

  return status;
}


/*---------------------------------------------------------------------------*/
/**	
 * Returns the trace file pointer for the db connection. 
 * If dbp is NULL, we return the global trace file pointer,
 * else the connection specific trace file pointer is returned.
 *
 * @param dbp           I - The database pointer.
 *
 * @return <ul>
 * <li>Connection specific or global trace file.
 * <li>NULL, if no trace file is open.
 * </ul>
 */
static FILE * 
DEFUN(_get_trace_fp, (dbp), const_sqlo_db_ptr_t dbp)
{
  FILE *fp;
  if (!dbp)
    fp =  _trace_fp;
  else {
    if (dbp->trace_fp) 
      fp = dbp->trace_fp;
    else
      fp = _trace_fp;
  }

  if (!fp)
    fp = stderr;

#ifndef NDEBUG
  fflush(fp);
#endif
  return (fp);

}



/*---------------------------------------------------------------------------*/
/**
 * Prepares a statement and returns the statement type.
 * If stmt_type is NULL, the statement type is not determined.
 * @param stp       I - The statement pointer
 * @param stmt      I - The sql statement
 * @param stmt_type O - The statement type, if stmt_type is not NULL
 * 
 * @return SQLO_SUCCESS or OCI status.
 */
static int
DEFUN(_prepare, (stp, stmt, stmt_type),
      const_sqlo_stmt_ptr_t stp AND
      const char * stmt AND
      ub2 * stmt_type)
{
  int status;

  status = OCIStmtPrepare(stp->stmthp, stp->dbp->errhp, (text *)stmt, 
                          strlen(stmt), OCI_NTV_SYNTAX, OCI_DEFAULT);

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "_prepare: OCIStmtPrepare: status=%d",
                   status););
  CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_prepare", (char*)stmt);

  if (stmt_type) {
    ub2 stype;

    /* Identify the statement type */
    status = OCIAttrGet((dvoid*) stp->stmthp, (ub4) OCI_HTYPE_STMT, 
                        (dvoid*) &stype,(ub4 *) 0, 
                        (ub4) OCI_ATTR_STMT_TYPE, 
                        (OCIError *) stp->dbp->errhp );

    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "_prepare", "OCIAttrGet");
   
    TRACE(2, fprintf(_get_trace_fp(stp->dbp), "_prepare: Statement type (%u): %s\n", 
                     (unsigned int) stype, _get_stmt_type_str( (int) stype) ););

    *stmt_type = stype;
  }

  return (SQLO_SUCCESS);
}

# if 0
/* Doesn't work currently. Will try it later again */
static sb4 _failover_callback_fn(dvoid * svchp, dvoid *envhp, dvoid *fo_ctx, 
                                 ub4 fo_type, ub4 fo_event);

static sb4 
DEFUN(_failover_callback_fn, (sbchp, envhp, fo_ctx, fo_tpe, fo_event),
      dvoid      * svchp AND
      dvoid      * envhp AND
      dvoid      * fo_ctx AND
      ub4        fo_type AND
      ub4        fo_event )
{
  sqlo_db_ptr_t dbp;

  dbp = (sqlo_db_ptr_t) fo_ctx;
  if (NULL == dbp) {
    TRACE(1, fprintf(NULL, "Failover handler received dbp == NULL !!!!\n"););
    return OCI_ERROR;
  }

  switch (fo_event) {
  case OCI_FO_BEGIN:
    EXEC_WHEN_THREADING( _stmtv_lock(); _dbv_lock(););
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failing over on dbh %d. Please stand by\n",
                     dbp->dbh););
    break;
  case OCI_FO_END:
    EXEC_WHEN_THREADING( _stmtv_unlock(); _dbv_unlock(););
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failover on dbh %d ended. Resuming services\n",
                     dbp->dbh););
    break;
  case OCI_FO_ABORT:
    EXEC_WHEN_THREADING( _stmtv_unlock(); _dbv_unlock(););
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failover on dbh %d aborted. No failover possible\n",
                     dbp->dbh););
    break;
  case OCI_FO_ERROR:
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failover on dbh %d had an error. Sleeping...\n",
                     dbp->dbh););
    sleep(3);
    TRACE(1, fprintf(_get_trace_fp(dbp), "Retrying.\n"););
    return(OCI_FO_RETRY);

    break;
  case OCI_FO_REAUTH:
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failover reauth on dbh %d. \n",
                     dbp->dbh););
    break;
  default:
    TRACE(1, fprintf(_get_trace_fp(dbp), "Failover: bad event %d\n", fo_event););
    break;
  }
  return(OCI_SUCCESS);
}

static int _register_callback __P(( sqlo_db_ptr_t dbp ));

static int 
DEFUN(_register_callback, (dbp), sqlo_db_ptr_t dbp)
{
  int status;
  OCIFocbkStruct failover;                 /*  failover callback structure */

  TRACE(1, fprintf(_get_trace_fp(dbp), "_register_callback starts\n"););
  /* initialize the context. */
  failover.fo_ctx = dbp;        /* This is passed as fo_ctx to the callback fn */

  failover.callback_function = &_failover_callback_fn;

  /* do the registration */
  status = OCIAttrSet(dbp->srvhp, (ub4) OCI_HTYPE_SERVER,
                      (dvoid *) &failover, (ub4) 0,
                      (ub4) OCI_ATTR_FOCBK, dbp->errhp);
  CHECK_OCI_STATUS_RETURN(dbp, status, "_register_callback", "");
  /* successful conclusion */
  return (SQLO_SUCCESS);
}
#endif
/*-------------------------------------------------------------------------
 * GLOBAL FUNCTIONS
 *-----------------------------------------------------------------------*/

/*-------------------------------------------------------------------------
 * sqlo_get_stmt
 *------------------------------------------------------------------------*/
const char *
DEFUN(sqlo_get_stmt, (sth), sqlo_stmt_handle_t sth )
{
  sqlo_stmt_ptr_t stp;
  CHECK_STHANDLE(stp, sth, "sqlo_get_stmt", NULL);

  if (_get_stmt_string(stp))
    return stp->stmt;

  return NULL;

}


/*-------------------------------------------------------------------------
 * sqlo_get_stmt_state
 *------------------------------------------------------------------------*/
int
DEFUN(sqlo_get_stmt_state, (sth), 
      sqlo_stmt_handle_t sth
     )
{
  sqlo_stmt_ptr_t stp;
  CHECK_STHANDLE(stp, sth, "sqlo_get_stmt_state", SQLO_INVALID_STMT_HANDLE);

  return _get_stmt_state(stp);
}


/*-------------------------------------------------------------------------
 * sqlo_init
 *------------------------------------------------------------------------*/
int
DEFUN(sqlo_init, (threaded_mode, max_db, max_cursors), 
      int threaded_mode AND 
      unsigned max_db AND 
      unsigned int max_cursors)
{
  int status = OCI_SUCCESS;

  
  if (threaded_mode)
    _oci_init_mode = OCI_THREADED;
  else
    _oci_init_mode = OCI_DEFAULT;

  /* get the environement settings */
  if (_sqlo_getenv())
    return(-1);

  if (TRACE_ENABLED && _trace_level > 0)
    _open_trace_file(NULL);

  /* Initialize OCI library */

  /* Oracle 8.0 OCI is initialized different */
#ifndef HAVE_OCIENVCREATE
  {
    ub4 mode = (ub4) _oci_init_mode;
    if (OCI_SUCCESS != (status = OCIInitialize(mode, 0, 0, 0, 0)))
      return(status);
  }
#endif

  if (max_db > 0) {
    _dbv_size = max_db ;
    _stmtv_size = max_cursors;
  }

  if (THREADS_ENABLED && OCI_THREADED == _oci_init_mode) {
#ifdef ENABLE_ORATHREADS
#  ifdef HAVE_OCIENVCREATE
  status = OCIEnvCreate(&_oci_envhp, _oci_init_mode, NULL, NULL, NULL, NULL, 0, NULL) ;
  if (status)
    return (status);
#  else
  if ( (status = OCIEnvInit((dvoid *)&_oci_envhp, _oci_init_mode, 0, (dvoid **)0)))
    return (status);
#  endif
  /* Alloc the global handle */
  if ((status = OCIHandleAlloc((dvoid *)_oci_envhp, (dvoid **)&_oci_errhp,
                               OCI_HTYPE_ERROR, 0, (dvoid **) 0)))
    return (status);
#endif
    if ((status = _init_mutexes()))
      return status;
  }

  /* Allocate the arrays for the pointers to the db structures */
  /* check if it the first time we are called */
  if (!_dbv) {
    /* We need to allocate the array of pointers to sqlo_db_t */
    _dbv = MALLOC(_dbv_size * sizeof(sqlo_db_ptr_t) );
    if (!_dbv) {
      return(SQLO_ERRMALLOC);
    }
    TRACE(4, fprintf(_get_trace_fp(NULL),
                     "sqlo_init: Allocated %u bytes for db handles\n", 
                     _dbv_size * (unsigned int)sizeof(sqlo_db_t)););
    memset(_dbv, 0, _dbv_size * sizeof(sqlo_db_ptr_t) );
  }

  /* Allocate the arrays for the pointers to the stmt structures */

  /* check if it the first time we are called */
  if (! _stmtv) {
    /* We need to allocate the array of pointers to sqlo_db_t */
    _stmtv = MALLOC (_stmtv_size * sizeof(sqlo_stmt_ptr_t )) ;
    if (! _stmtv) {
      return (SQLO_ERRMALLOC) ;
    }
      TRACE(4, fprintf(_get_trace_fp(NULL),
                       "sqlo_init_add: Allocated %u bytes for stmt handles\n",
                       _stmtv_size * (unsigned int) sizeof(sqlo_stmt_t*)););
      memset(_stmtv,0, _stmtv_size * sizeof(sqlo_stmt_ptr_t )) ;
  }

  _sqlo_init = TRUE;
  return (status);
}


/*---------------------------------------------------------------------------
 *	 sqlo_trace
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_trace, (dbh, on), sqlo_db_handle_t dbh AND int on)
{
  int stat;
  if (on) 
    stat = sqlo_exec(dbh, "ALTER SESSION SET SQL_TRACE TRUE");
  else 
    stat = sqlo_exec(dbh, "ALTER SESSION SET SQL_TRACE FALSE");

  return stat;
}


/*---------------------------------------------------------------------------
 *	 sqlo_geterror
 *--------------------------------------------------------------------------*/
const char * 
DEFUN(sqlo_geterror, (dbh), sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;

                               /* An invalid handle may be passed to us, if
                                * the connect failed. In this case we return
                                * the global error message.
                                */
  if (dbh >= (int) _dbv_size || dbh < 0 || !_dbv[ dbh ]->used) {
    TRACE(1, fprintf(_get_trace_fp(NULL), 
                     "Invalid Database handle %d in sqlo_geterror. Returning _errmsg\n",
                     dbh););
    return _errmsg;
  }  else {
    dbp = _dbv[dbh];
    
    if (!dbp)
      return "Invalid database handle in call to sqlo_geterror";
    
    return (const char *) dbp->errmsg;
  }

}


/*---------------------------------------------------------------------------
 * sqlo_geterrcode
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_geterrcode, (dbh), sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_geterrcode", SQLO_INVALID_DB_HANDLE);

  return(_get_errcode(dbp));

}


/*---------------------------------------------------------------------------
 *	 sqlo_exists
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_exists, (dbh, table, colname, colval, where),
       sqlo_db_handle_t dbh AND
       const char * table AND
       const char * colname AND
       const char * colval AND
       const char * where)
{
  char stmt[4096];
  int argc = 0;
  const char * argv[1];
  sqlo_stmt_handle_t sth = SQLO_STH_INIT;
  int status = 0;
  int retcode = SQLO_ERROR;

  sqlo_db_ptr_t  dbp;
  
  CHECK_DBHANDLE(dbp, dbh, "sqlo_exists", SQLO_INVALID_DB_HANDLE);

  if ( !table || *table == '\0' ) {
    sprintf(dbp->errmsg,"No table specified for sqlo_exists\n");
    TRACE(1, (void) fputs(dbp->errmsg, _get_trace_fp(dbp)); );
    return (SQLO_ERROR);
  }

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_exists: on %s\n", table););

  if (colval && *colval != '\0') {
    argv[0] = colval;
    argc = 1;
    sprintf(stmt, "select 'YES' from %s where %s = :b1",table, colname);
    if (where && *where != '\0') {
      strcat( stmt," and "); strcat (stmt, where);
    }
  } else {
    sprintf ( stmt, "select 'YES' from %s", table);
    if (where && *where != '\0') {
      strcat(stmt, " where "); strcat (stmt, where );
    }
  }

  TRACE(2, fprintf( _get_trace_fp(dbp), "sqlo_exists: %s\n",stmt););

  while (SQLO_STILL_EXECUTING == (status = sqlo_open2(&sth, dbh, stmt, argc, argv))) {
    TRACE(2, fprintf( _get_trace_fp(dbp), "sqlo_exists: sqlo_open2 status=%d sth=%d\n",
                      status, sth););
    SQLO_USLEEP;
  }

  if (status < 0) {
      CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exists", "sqlo_open2");
    }

  while (SQLO_STILL_EXECUTING == (status = sqlo_fetch(sth, 1)))  {
    SQLO_USLEEP;
  }

  if (status < 0) {
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exists", "sqlo_fetch");
  } else if (status == 0) {
    retcode = SQLO_SUCCESS;           /* exists */
  } else {
    retcode = SQLO_NO_DATA;  /* not exists */
  }

  status = sqlo_close(sth);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exists", "sqlo_close");

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_exists: %s (%d)\n", retcode ? "NO" : "YES",
                   retcode););

  return retcode;
}


/*---------------------------------------------------------------------------
 *	 sqlo_count
 *--------------------------------------------------------------------------*/
int
DEFUN(sqlo_count, (dbh, table, colname, colval, where),
       sqlo_db_handle_t dbh AND
       const char * table AND
       const char * colname AND
       const char * colval AND
       const char * where)
{
  char stmt[4096];
  int cnt = SQLO_ERROR;
  int argc = 0;
  const char * argv[1];
  const char **v;
  sqlo_stmt_handle_t sth = -1;
  int status;
  sqlo_db_ptr_t  dbp;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_count", SQLO_INVALID_DB_HANDLE);

  if ( !table || *table == '\0' ) {
    sprintf(dbp->errmsg, "No table specified for sqlo_count\n");
    TRACE(1, (void) fputs(dbp->errmsg, _get_trace_fp(dbp)); );
    return (SQLO_ERROR);
  }

  if ( colname && *colname != '\0' )
    sprintf (stmt, "select count (*) from %s ", table);
  else
    sprintf (stmt, "select count (*) from %s ", table);

  if (colval && *colval != '\0') {
    argv[0] = colval;
    argc = 1;
    sprintf ( &stmt[strlen ( stmt )], "where %s = :b1", colname);
    if (where && *where != '\0') {
      strcat(stmt, " and "); strcat (stmt, where);
    }
  } else {
    if (where && *where != '\0') {
      strcat( stmt, " where "); strcat (stmt, where);
    }
  }
  
  while (SQLO_STILL_EXECUTING == (status = sqlo_open2(&sth, dbh, stmt, argc, argv)))  {
      SQLO_USLEEP;
    }

  if (status < 0) {
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_count", "sqlo_open2");
  }

  while (SQLO_STILL_EXECUTING == (status = sqlo_fetch(sth, 1)))  {
      SQLO_USLEEP;
    }

  if (status < 0) {
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_count", "sqlo_fetch");
  } else if (status == 0) {
    v = sqlo_values(sth, NULL, 1);
    cnt = atoi(*v);
  } else {
    cnt = 0;
  }

  status = sqlo_close(sth);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_count", "sqlo_close");

  return ( cnt );
}


/*---------------------------------------------------------------------------
 * sqlo_run 
 *--------------------------------------------------------------------------*/
int
DEFUN(sqlo_run, (dbh, stmt, argc, argv),
       sqlo_db_handle_t dbh AND const char * stmt AND int argc AND const char **argv )
{
  sqlo_db_ptr_t  dbp;
  int ret;
  int status;
  sqlo_stmt_handle_t sth = -1;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_run", SQLO_INVALID_DB_HANDLE);

  TRACE(2, 
  { 
    int z ; 
    for ( z=0 ; z < argc; z++) { 
      fprintf(_get_trace_fp(dbp), "arg[%02d]: %s\n", z, argv[z] ? argv[z] : "NULL");
    } 
  });

  while (SQLO_STILL_EXECUTING == (status = sqlo_open2(&sth, dbh, stmt, argc, argv)))  {
    SQLO_USLEEP;
  }

  if (status < 0) 
    return status;

  while (SQLO_STILL_EXECUTING == (ret = sqlo_fetch(sth, 1))) {
    SQLO_USLEEP;
  }

  if (0 <= ret )
	ret = sqlo_prows(sth);

  sqlo_close(sth);

  return ret;
}


/*---------------------------------------------------------------------------
 *	 sqlo_exec 
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_exec, (dbh, stmt), sqlo_db_handle_t dbh AND const char * stmt)
{
  sqlo_db_ptr_t  dbp;

  int status, status2;
  ub4 prows;


  CHECK_DBHANDLE(dbp, dbh, "sqlo_exec", SQLO_INVALID_DB_HANDLE);

  /* we can use a local variable, because we will never return a 
   * SQLO_OCI_STILL_EXECUTING
   */
  /* Use the stmthp in dbp. */
  if (dbp->stmthp == NULL) {
    TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_exec: prepare: %s\n", stmt););

    status = OCIHandleAlloc( (dvoid *) dbp->envhp, (dvoid **) &dbp->stmthp,
                             OCI_HTYPE_STMT, (size_t) 0, (dvoid **) 0);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exec", "OCIHandleAlloc(stmt)");
    
    
    status = OCIStmtPrepare(dbp->stmthp, dbp->errhp, (text *)stmt, 
                            strlen(stmt), OCI_NTV_SYNTAX, OCI_DEFAULT);
    
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exec(prepare)", (char*)stmt);
  } else {
    /* still executing */
    ;
  }
  
  status = OCIStmtExecute(dbp->svchp, 
                          dbp->stmthp, 
                          dbp->errhp, 
                          (ub4) 1, 
                          (ub4) 0,
                          ( OCISnapshot *) 0, 
                          (OCISnapshot *) 0,
                          (ub4) OCI_DEFAULT);

  if (status == SQLO_SUCCESS) {

    /* finished the call. Get row count */
    status = OCIAttrGet((dvoid*)dbp->stmthp,
                        (ub4) OCI_HTYPE_STMT, 
                        (dvoid *)  &prows, 
                        (ub4 *) 0, 
                        (ub4) OCI_ATTR_ROW_COUNT, 
                        dbp->errhp);

    if (status != 0) {
      (void)OCIHandleFree(dbp->stmthp, OCI_HTYPE_STMT);

      CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_exec", "get row_count");
    }

    status = prows;

    status2 = OCIHandleFree(dbp->stmthp, OCI_HTYPE_STMT);
    CHECK_OCI_STATUS_RETURN(dbp, status2, "sqlo_exec", "OCIHandleFree(1)");

    dbp->stmthp = 0;
    dbp->status = SQLO_SUCCESS;

  }  else if (status == OCI_STILL_EXECUTING) {

    dbp->status = SQLO_STILL_EXECUTING;
    return (SQLO_STILL_EXECUTING);

  } else {

    dbp->status = status;
    _save_oci_status(dbp, "sqlo_exec", stmt, __LINE__);
    status2 = OCIHandleFree(dbp->stmthp, OCI_HTYPE_STMT); 
    dbp->stmthp = 0;
    CHECK_OCI_STATUS_RETURN(dbp, status2, "sqlo_exec", "OCIHandleFree(2)");
  }

  return (status);
}


/*---------------------------------------------------------------------------
 * sqlo_open
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_open,(dbh, stmt, argc, argv),
       sqlo_db_handle_t dbh AND const char * stmt AND
       int argc AND
       const char ** argv) 
{
  sqlo_db_ptr_t  dbp;
  sqlo_stmt_ptr_t stp;
  int status;
  int ret;
  bool_t bmf = FALSE;                   /* flag indicates change in blocking mode */
  unsigned int blocking;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_open", SQLO_INVALID_DB_HANDLE);

  ret = SQLO_SUCCESS;

  /* if we are in non-blocking mode, we switch back to blocking */
  status = _get_blocking_mode(dbp, &blocking);
  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "sqlo_set_blocking_i");
    
  if (SQLO_OFF == blocking) {
    status = sqlo_set_blocking(dbh, 1);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "sqlo_set_blocking(dbh, 1)");
    bmf = TRUE;
  }

  /* Get a statement handle pointer */
  status = _stmt_new(dbp, stmt, &stp);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "_stmt_new");

  TRACE(2, fprintf(_get_trace_fp(dbp), "Parse: %s\n", stmt););

  /* prepare the statement and determine the statement type */
  status = _prepare(stp, stmt, &(stp->stype));
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "prepare");

  /* We cannot handle PL/SQL blocks here */
  if (OCI_STMT_BEGIN  == stp->stype ||
      OCI_STMT_DECLARE == stp->stype) {
    status = sqlo_close( (int) stp->sth);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "sqlo_close, ERROR: INVALID STMT TYPE");
    return(SQLO_INVALID_STMT_TYPE);
  }

  stp->prepared = TRUE;
  
  if (stp->stype == OCI_STMT_SELECT) {

    /* Set the prefetch count to _num_prefetch_rows */
    status = _set_prefetch_rows(stp, _num_prefetch_rows);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "_set_prefetch_rows");
  
  }

  ret = sqlo_reopen((int)stp->sth, argc, argv);
  /* Dont check the retcode again. Errors are already saved during reopen */
  
  if ( SQLO_SUCCESS == ret ) {
    stp->opened = TRUE;
  }

  /* switch back to non-blocking mode */
  if (bmf) {
    status = sqlo_set_blocking(dbh, 0 );
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open", "sqlo_set_blocking(dbh, 0)");

  }
  return( (SQLO_SUCCESS == ret) ? (int) stp->sth : ret);

}


/*---------------------------------------------------------------------------
 * sqlo_open2 
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_open2,(sthp, dbh, stmt, argc, argv),
      sqlo_stmt_handle_t * sthp AND 
      sqlo_db_handle_t dbh AND 
      const char * stmt AND
      int argc AND
      const char ** argv) 
{
  sqlo_db_ptr_t  dbp;
  sqlo_stmt_ptr_t stp = NULL;
  int status;
  int ret;
  unsigned int blocking;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_open2", SQLO_INVALID_DB_HANDLE);
  if (!sthp)
    return SQLO_INVALID_STMT_HANDLE;

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_open2: dbh=%d sth=%d\n", dbh, *sthp););

  if (!sthp)
    return(SQLO_INVALID_STMT_HANDLE);

  /* check if the statement handle passed to us is valid, if we
   * are in non-blocking mode.*/
  status = _get_blocking_mode(dbp, &blocking);

  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open2", "_get_blocking_mode");

  if ( SQLO_OFF == blocking ) {
    sqlo_stmt_handle_t sth = *sthp;
    stp = _stmtv[sth];
    if ( sth >= 0 && 
         sth < (int)_stmtv_size &&
         stp &&
         stp->used &&
         stp->opened && 
         stp->stype == OCI_STMT_SELECT &&
         stp->still_executing) {
      return(sqlo_reopen(sth, argc, argv));
    }
  }

  /* Get a statement handle pointer */
  status = _stmt_new(dbp, stmt, &stp);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open2", "_stmt_new");

  TRACE(2, fprintf(_get_trace_fp(dbp), "Parse: %s\n", stmt););

  /* prepare the statement and determine the statement type */
  status = _prepare(stp, stmt, &(stp->stype));
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open2", "_prepare");

  /* We cannot handle PL/SQL blocks here */
  if (OCI_STMT_BEGIN  == stp->stype ||
      OCI_STMT_DECLARE == stp->stype) {
    status = sqlo_close((int)stp->sth);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open2", "sqlo_close, ERROR: INVALID STMT TYPE");
    return (SQLO_INVALID_STMT_TYPE);
  }


  stp->prepared = TRUE;
  
  if (stp->stype == OCI_STMT_SELECT) {
    /* Set the prefetch count to _num_prefetch_rows */
    status = _set_prefetch_rows(stp, _num_prefetch_rows);

    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_open2", "_set_prefetch_rows");
  
  }

  ret = sqlo_reopen((int)stp->sth, argc, argv);
  /* Dont check the retcode again. Errors are already saved during reopen */
  
  if ( SQLO_SUCCESS == ret || SQLO_STILL_EXECUTING == ret) {
    stp->opened = TRUE;
    *sthp = (int) stp->sth;
  }

  return (ret);

}


/*---------------------------------------------------------------------------
 * sqlo_reopen
 *-------------------------------------------------------------------------*/
int
DEFUN(sqlo_reopen,  (sth, argc, argv),
       sqlo_stmt_handle_t sth AND 
       int argc AND 
       const char ** argv)
{
  register sqlo_stmt_ptr_t  stp;
  int status = SQLO_SUCCESS;

  CHECK_STHANDLE(stp, sth, "sqlo_reopen", SQLO_INVALID_STMT_HANDLE);

  if (argc > 0) {

    if (!stp->still_executing) {
      /* Reset number of bindpv. Don't free them! */
      _bindpv_reset(stp);

      status = _bind_argv(stp, (unsigned int)argc, argv);

      if (SQLO_SUCCESS != status) {
        return (status);
      }
            
      TRACE(2, 
            if (argc) {
              fprintf(_get_trace_fp(stp->dbp),"(Re)open [%2d] %s\n", sth, _get_stmt_string(stp) );
            });

      TRACE(3, { 
        int z ; 
        for ( z=0 ; z < argc; z++) { 
          fprintf(_get_trace_fp(stp->dbp), "arg[%02d]: %s\n", z, argv[z] ? argv[z] : "NULL");
        } });
    }
  }

  if (stp->stype == OCI_STMT_SELECT ) {

    if (0 == stp->num_executions) 
      status = _define_output(stp);
    else {
      /* execute the cursor to open the statement again */
      if (OCI_STILL_EXECUTING == 
          (status = OCIStmtExecute(stp->dbp->svchp, 
                                   stp->stmthp, 
                                   stp->dbp->errhp, 
                                   (ub4) 0, 
                                   (ub4) 0, 
                                   (OCISnapshot *) 0, 
                                   (OCISnapshot *) 0,
                                   (ub4) OCI_DEFAULT)
           )
          )
        {
          stp->still_executing = TRUE;
          stp->dbp->status = SQLO_STILL_EXECUTING;
          return status;
        } else {
          CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_reopen", "OCIStmtExecute");
        }
    }

    if (OCI_SUCCESS != status) {
      if (OCI_STILL_EXECUTING == status) {
        stp->still_executing = TRUE;
        stp->dbp->status = SQLO_STILL_EXECUTING;
        return (SQLO_STILL_EXECUTING);
      } else {
        stp->dbp->status = status;
        _save_oci_status(stp->dbp, "sqlo_reopen", "OCIStmtExecute(query)",
                    __LINE__);
        return(status);
      }
    } else {
      stp->dbp->status = SQLO_SUCCESS;
      stp->still_executing = 0;
      if (0 == stp->num_executions ) {
        stp->num_executions++;
        status = _define_output(stp);
      }
    }
  }

  return (status);
}


/*---------------------------------------------------------------------------
 * sqlo_fetch
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_fetch, (sth, nrows), sqlo_stmt_handle_t sth AND unsigned int nrows)
{
  sqlo_stmt_ptr_t  stp;
  sqlo_db_ptr_t  dbp;
  int status;

  CHECK_STHANDLE(stp, sth, "sqlo_fetch", SQLO_INVALID_STMT_HANDLE);

  if (THREADS_ENABLED && OCI_THREADED == _oci_init_mode &&  
      TRUE != _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
    sprintf(stp->dbp->errmsg,
            "sqlo_fetch [%2d]: Try to fetch from a cursor opened by another thread !!!\n", 
            sth);
    TRACE(1, fprintf(_get_trace_fp(stp->dbp), stp->dbp->errmsg););
    return SQLO_INVALID_STMT_HANDLE;
  }


  dbp = stp->dbp;
  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_fetch sth=%d, nrows=%u\n", sth, nrows););


  if (stp->stype != OCI_STMT_SELECT)  {
    if (!stp->still_executing) {
      TRACE(2, fprintf(_get_trace_fp(dbp), "Exec [%2d] %.65s\n", sth, _get_stmt_string(stp)););
    }

    status = OCIStmtExecute(dbp->svchp, stp->stmthp, dbp->errhp, 
                            (ub4) 1, (ub4)0,
                            ( OCISnapshot *) 0, (OCISnapshot *) 0,
                            (ub4) OCI_DEFAULT);

    if (OCI_SUCCESS != status) {
      if (OCI_STILL_EXECUTING == status) {
        stp->still_executing = TRUE;
        status = dbp->status = SQLO_STILL_EXECUTING;
      } else {
        dbp->status = status;
        _save_oci_status(dbp,"sqlo_fetch", "OCIStmtExecute(nonquery)", __LINE__);
      }
    } else {
      stp->still_executing = 0;
      stp->dbp->status = SQLO_SUCCESS;
    }
  } else {
#ifdef HAVE_OCISTMTFETCH2
    status = OCIStmtFetch2(stp->stmthp, dbp->errhp, nrows, 
                          OCI_FETCH_NEXT, 0, OCI_DEFAULT);
#else    
    /* deprecated in new versions of Oracle (9i) */
    status = OCIStmtFetch(stp->stmthp, dbp->errhp, nrows,
                          OCI_FETCH_NEXT, OCI_DEFAULT);
#endif
    TRACE(3, fprintf(_get_trace_fp(dbp), "OCIStmtFetch finished with %d\n", status););

    if (status != OCI_SUCCESS && status != OCI_NO_DATA) {
      if (OCI_STILL_EXECUTING == status) {
        stp->still_executing = TRUE;
        return(SQLO_STILL_EXECUTING);
      }
      CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_fetch", "OCIStmtFetch");
    } else {
      stp->still_executing = 0;
      stp->dbp->status = status;
    }
  }

  return (status);

}


/*---------------------------------------------------------------------------
 * sqlo_values
 *-------------------------------------------------------------------------*/
const char ** 
DEFUN(sqlo_values, (sth, num, do_strip_string),
       sqlo_stmt_handle_t sth AND 
       int * num AND 
       int do_strip_string)
{
  register sqlo_col_ptr_t colp;
  register char ** outpp;
  register unsigned int col_idx;
  register ub2 * lenp;
  register ub2 * indp;
  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_values", NULL);

  if (THREADS_ENABLED && OCI_THREADED == _oci_init_mode && 
      TRUE != _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
    sprintf(stp->dbp->errmsg,
            "sqlo_values [%2d]: Try to get values of a cursor opened by another thread !!!\n", 
            sth);
    TRACE(1, fprintf(_get_trace_fp(stp->dbp), stp->dbp->errmsg););
    return NULL;
  }

  if (stp->stype != OCI_STMT_SELECT || !stp->opened) {
    sprintf(stp->dbp->errmsg, 
            "Cannot get values for a non-select/non-opened statement (sth %d) passed to sqlo_values\n", sth);
    TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
    if (num)
      *num = 0;
	return NULL;
  }

  colp  = stp->ocolsv;

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "Get values [%2d]. _strip_string: %d\n", 
                   sth, do_strip_string););

  /* 
   * Since each variable is returned as character string, very little 
   * processing is required here.
   */
  outpp = stp->outv;
  lenp  = stp->rlenv;
  indp  = stp->oindv;
  for (col_idx = 0; col_idx < stp->num_defnpv; 
       ++col_idx, ++colp, ++outpp, ++lenp, ++indp ) {

    if (*indp) { /* 0 indicates 'not null' -1 indicates 'null' */
      **outpp = '\0';
      *lenp = 0;
    } else {
      (*outpp)[ *lenp ] = '\0';
    }

    if (do_strip_string && *lenp) {
      if (colp->dtype != SQLT_NUM && colp->dtype != SQLT_LNG) {
        _strip_string(*outpp, *lenp);
      } else {
        /* Strip leading blanks for numbers */
        register char *p;
        unsigned int l = *lenp + 1;
        for (p = *outpp; *p == ' '; ++p) {
          ;
        }

        if (p != *outpp ) {
          memmove(*outpp, p, l); 
        }
      }
    }
    TRACE(3, fprintf(_get_trace_fp(stp->dbp), 
                     "  L: %2u argv[%02u]: %s\n", *lenp, col_idx, *outpp ););
  }

  if (num)
    *num = (int)stp->num_defnpv;

  return (const char **) stp->outv;
}


/*---------------------------------------------------------------------------
 *	sqlo_prows
 *-------------------------------------------------------------------------*/
int
DEFUN(sqlo_prows, (sth), sqlo_stmt_handle_t sth)
{
  sqlo_stmt_ptr_t  stp;
  int status;
  ub4 prows;

  CHECK_STHANDLE(stp, sth, "sqlo_prows", SQLO_INVALID_STMT_HANDLE);

  if ( !stp->opened ) {
    sprintf(stp->dbp->errmsg,
            "Cannot get processed rows for a non-executed/fetched statement "
            "(sth %d) passed to sqlo_prows\n", sth);
    TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
    return SQLO_INVALID_STMT_HANDLE;
    
  }

  
  status = OCIAttrGet((dvoid*)stp->stmthp,
                      (ub4) OCI_HTYPE_STMT, 
                      (dvoid *)  &prows, 
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_ROW_COUNT, 
                      stp->dbp->errhp);
      
  CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_prows(GetAttr)", "ROW_COUNT");

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "Get proc. rows [%2d]: %u\n", sth, 
                     prows););
  
  return( (int) prows);
}


/*---------------------------------------------------------------------------
 *	sqlo_ncols
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_ncols, (sth, in), sqlo_stmt_handle_t sth AND int in) 
{
  ub4 ncols;
  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_ncols", SQLO_INVALID_STMT_HANDLE);

  if (!stp->used ) {
    sprintf(stp->dbp->errmsg, "Invalid sth %d passed to sqlo_ncols\n", sth);
    TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
    return SQLO_INVALID_STMT_HANDLE;
  }

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "Get NCols [%2d] for %s desc\n",
                   sth, in ? "in" : "out" ););

  if (in)
      ncols = (ub4) stp->num_bindpv;
  else {
    int status;

    if (0 == stp->num_executions && ! (REFCURSOR == stp->cursor_type) ) {
      /* execute to describe the output */
      while (OCI_STILL_EXECUTING == 
             (status = OCIStmtExecute(stp->dbp->svchp, 
                                      stp->stmthp, 
                                      stp->dbp->errhp, 
                                      (ub4) 0, 
                                      (ub4) 0, 
                                      (OCISnapshot *) 0, 
                                      (OCISnapshot *) 0,
                                      (ub4) OCI_DEFAULT))) {
        SQLO_USLEEP;
      }
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_ncols", 
                            "OCIStmtExecute(DESCRIBE)");
      ++(stp->num_executions);
    }

    status = OCIAttrGet((dvoid*)stp->stmthp,
                        (ub4) OCI_HTYPE_STMT, 
                        (dvoid *)  &ncols, 
                        (ub4 *) 0, 
                        (ub4) OCI_ATTR_PARAM_COUNT, 
                        stp->dbp->errhp);
    
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_ncols", 
                            "OCIAttrGet(NumberOfColumns)");
  }

  TRACE(3, fprintf(_get_trace_fp(stp->dbp), "  NCols: %u\n", (unsigned int) ncols););

  return( (int)ncols);
}


/*---------------------------------------------------------------------------
 *	sqlo_command
 *-------------------------------------------------------------------------*/
const char * 
DEFUN(sqlo_command, (sth), sqlo_stmt_handle_t sth)
{
  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_command", NULL);
  return(_get_stmt_string(stp));

}


/*---------------------------------------------------------------------------
 *	sqlo_close
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_close, (sth), sqlo_stmt_handle_t sth)
{
  sqlo_stmt_ptr_t  stp;            /* Statement pointer */
  int status;                   /* OCI status */
  unsigned defnp_idx;           /* Loop index for defnp */

  CHECK_STHANDLE(stp, sth, "sqlo_close", SQLO_INVALID_STMT_HANDLE);


  /* make sure we close only cursors in our thread */
  EXEC_WHEN_THREADING
    (
     if ( TRUE != _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
       sprintf(stp->dbp->errmsg,
               "sqlo_close [%2d]: Try to close a cursor opened by another thread !!!\n", 
               sth);
       TRACE(1, fprintf(_get_trace_fp(stp->dbp), stp->dbp->errmsg););
       return SQLO_INVALID_STMT_HANDLE;
     }
     );

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "sqlo_close [%2d]: %.60s\n", sth, 
                   _get_stmt_string(stp)););

  TRACE(3, fprintf(_get_trace_fp(stp->dbp), "sqlo_close [%2d]: opened=%d, prepared=%d\n",
                   sth, stp->opened, stp->prepared););

  _bindpv_reset(stp);

  for (defnp_idx = 0; defnp_idx < stp->num_defnpv; ++defnp_idx) {

    /* deallocate memory allocated for LONG columns */
    if (stp->outv_size[ defnp_idx ] >= _max_long_size) {
      FREE(stp->outv[ defnp_idx ]);
      stp->outv[ defnp_idx ] = NULL;
      stp->outv_size[ defnp_idx ] = 0;
    }

    if (stp->ocolsv && stp->ocolsv[ defnp_idx ].paramd) {
      status = OCIDescriptorFree(stp->ocolsv[ defnp_idx ].paramd, OCI_DTYPE_PARAM);
      stp->ocolsv[ defnp_idx ].paramd = NULL;
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_close", "OCIDescriptorFree(paramd)");
    }
  } /* end for defnp_idx */
      
  memset(stp->defnpv, 0, stp->num_defnpv * sizeof(OCIDefine *));
  stp->num_defnpv = 0;

  if ( stp->stmthp ) {
    status = OCIHandleFree(stp->stmthp, OCI_HTYPE_STMT);
    stp->stmthp = NULL;
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_close", "OCIHandleFree(stmthp)");
  }

  stp->dbp->status = SQLO_SUCCESS;


  _stmt_release(stp);       /* make sth available to others. */

  return (SQLO_SUCCESS);
}


/*---------------------------------------------------------------------------
 *	sqlo_print
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_print, (sth), sqlo_stmt_handle_t sth )
{
  sqlo_stmt_ptr_t  stp;
  register unsigned int col_idx;

  CHECK_STHANDLE(stp, sth, "sqlo_print", SQLO_INVALID_STMT_HANDLE);

  printf("Stmt in sth %d: %s\n", sth, stp->stmt);

  if (!stp->opened) 
    printf(" not ");
  printf(" opened, ");

  if (!stp->prepared) 
    printf(" not ");
  printf(" prepared, ");
  
  printf("Stmt Type: %d (%s)\n", (int)stp->stype, _get_stmt_type_str((int)stp->stype));
  printf("N-Bindpv: %u, (allocated: %u)\n", stp->num_bindpv, stp->bindpv_size);
  printf("prows: %d\n", sqlo_prows(sth));
  

  if (stp->stype == OCI_STMT_SELECT) {
    for (col_idx = 0; col_idx < stp->num_defnpv; ++col_idx) {
      printf("Colum Name[%02u]  : %s\n", col_idx, stp->ocolsv[ col_idx ].col_name);
      printf("Datatype        : %d (%s)\n", (int)stp->ocolsv[ col_idx ].dtype, 
             _get_data_type_str((int)stp->ocolsv[ col_idx ].dtype));
      printf("Buffer size     : %u\n", stp->outv_size[ col_idx ]);
      printf("DB size         : %d\n", (int) stp->ocolsv[ col_idx ].dbsize);
      printf("Prec            : %d\n", (int) stp->ocolsv[ col_idx ].prec);
      printf("Scale           : %d\n", (int) stp->ocolsv[ col_idx ].scale);
      printf("NullOk          : %d\n", (int) stp->ocolsv[ col_idx ].nullok);
    }

  }

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 *	sqlo_split_cstring
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_split_cstring, (cstr, uid, pwd, tnsname, bufsize), 
      const char * cstr AND char * uid AND char * pwd AND char * tnsname AND
      unsigned int bufsize)
{
  char *c = (char*) cstr;
  unsigned int n;
  char * env_ptr;

  if (!cstr || !uid || !pwd || !tnsname || !bufsize)
    return (SQLO_ERROR);

  /* extract username, password and tnsname from the connect string */
  *uid = '\0';
  *pwd = '\0';
  *tnsname = '\0';

  /* copy username part to uid */
  n = 0;
  while ( *c && *c !='/' ) {
    if (n >=  bufsize )
      return SQLO_ERROR;
    *(uid++) = *(c++);
    ++n;
  }

  *uid = '\0';
 
  /* copy password part, if present */
  if ( *c == '/') {
    ++c;
    n = 0;
    while( *c && *c != '@')  {
      if (n >= bufsize)
        return SQLO_ERROR;
      *(pwd++) = *(c++);
      ++n;
    }
    *pwd = '\0';
  }

  /* copy tnsname if present, otherwise we use oracle sid */
  if( *c == '@' )  {
    ++c;
    n = 0;
    while ( *c != '\0' ) {
      if (n >= bufsize)
        return SQLO_ERROR;
      *(tnsname++) = *(c++);
      ++n;
    }
    *tnsname = '\0';
  } else {
    if ( (env_ptr = getenv("ORACLE_SID")) ) {
      strncpy(tnsname, env_ptr, bufsize-1);
      tnsname[bufsize-1] = '\0'; 
    }
  }
  return SQLO_SUCCESS;

}


/*---------------------------------------------------------------------------
 *	sqlo_server_attach 
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_server_attach, (dbhp, tnsname), 
      sqlo_db_handle_t * dbhp AND 
      const char * tnsname)
{
  int status;
  sqlo_db_ptr_t  dbp;
  
  if (dbhp)
    *dbhp = -1;

  TRACE(2, fprintf(_get_trace_fp(NULL), "sqlo_server_attach starts\n" ););

  if (!(dbp = _db_add())) {
    return SQLO_ERROR;
  }

  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_server_attach: allocated dbh=%u\n", 
                   dbp->dbh); );

  /* if tnsname contains a connect string we split it */
  if (strchr(tnsname, '@') || strchr(tnsname, '/') ) {
    enum {MAX_BUFLEN = 1023};
    char uid[MAX_BUFLEN + 1];
    char pwd[MAX_BUFLEN + 1];
    char dbname[MAX_BUFLEN + 1];

    if (SQLO_SUCCESS != sqlo_split_cstring(tnsname, uid, pwd, dbname, MAX_BUFLEN))
      return SQLO_ERROR;
    dbp->tnsname = strdup(dbname);
    
  } else {
    dbp->tnsname = strdup(tnsname);
  }

  dbp->status = SQLO_SUCCESS;
  dbp->errmsg[0] = '\0';

  /* We return the handle even on error, so sql_geterror can return 
   * the stored message
   */
  if (dbhp)
    *dbhp = (int) dbp->dbh;

  /*
    We must use OCIEnvCreate instead of OCIInitialize / OCIEnvInit (see Oracle OCI
    Programmer's Guide 8.1.6 pg 15-91 - 15-93). If during the first call to OCIEnvCreate
    an OCI_THREADED parameter is specified then this parameter will apply all along.
    CAFL says fb doesn't have the earlier manual...which says to use
    ocienvinit to create the handle.  The real question is whether to pass
    OCI_NO_MUTEX to it, which would probably be right here where there is one
    environment handle per connection, or OCI_DEFAULT.  I have no guts.
  */

  /* There are roumors that OCIEnvCreate is not thread safe. I'm not sure that this
   * is true, but to be on the safe side, I protect it by a lock
   */
  EXEC_WHEN_THREADING(_env_lock(););

#ifdef HAVE_OCIENVCREATE
  status = OCIEnvCreate(&dbp->envhp, _oci_init_mode, NULL, NULL, NULL, NULL, 0, NULL) ;
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIEnvCreate" );
#else
  status = OCIEnvInit((dvoid *)&dbp->envhp, _oci_init_mode, 0, (dvoid **)0);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIEnvInit" );
#endif

  EXEC_WHEN_THREADING(_env_unlock(););

  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Alloc the service context handle */
  status = OCIHandleAlloc((dvoid *)dbp->envhp, (dvoid **)&dbp->svchp,
                           OCI_HTYPE_SVCCTX, 0, (dvoid **) 0);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIEnvHandleAlloc(svchp)" );

  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Alloc the error handle */
  status = OCIHandleAlloc((dvoid *)dbp->envhp, (dvoid **)&dbp->errhp,
                           OCI_HTYPE_ERROR, 0, (dvoid **) 0);

  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIHandleAlloc(errhp)" );
  
  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Alloc the server handle */
  status = OCIHandleAlloc ((dvoid *)dbp->envhp, (dvoid **)&dbp->srvhp,
                           OCI_HTYPE_SERVER, 0, (dvoid **) 0);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIHandleAlloc(srvhp)" );

  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Create a server context */
  status = OCIServerAttach (dbp->srvhp, dbp->errhp, (text *)dbp->tnsname, 
                            strlen(dbp->tnsname), OCI_DEFAULT);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCISeverAttach(tnsname)" );

  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Set the server context in the service context */
  status = OCIAttrSet ((dvoid *)dbp->svchp, OCI_HTYPE_SVCCTX, 
                       (dvoid *)dbp->srvhp, (ub4) 0, OCI_ATTR_SERVER, 
                       dbp->errhp);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIAttrSet(server->service)" );
  
  /* release allocated resources */
  if (OCI_SUCCESS != status) {
    _db_release(dbp);
    return (status);
  }

  /* Allocate a authentication handle */
  status = OCIHandleAlloc ((dvoid *)dbp->envhp, (dvoid **)&dbp->authp,
                           OCI_HTYPE_SESSION, 0, (dvoid **) 0);
  CHECK_OCI_STATUS(dbp, status, "sqlo_server_attach", "OCIHandleAlloc(authp)" );

  if (SQLO_SUCCESS == status) {
    TRACE(2,fprintf(_get_trace_fp(dbp), "sqlo_server_attach[%d]: attached\n", *dbhp); );
  } else {
    /* release allocated resources */
    TRACE(2,fprintf(_get_trace_fp(dbp), "sqlo_server_attach[%d]: failed\n", *dbhp); );
    _db_release(dbp);
    return (status);
  }
  dbp->used   = TRUE;
  dbp->attached = TRUE;           

#if 0
  _register_callback(dbp);
#endif
  return (status); 
}


/*---------------------------------------------------------------------------
 *	sqlo_session_begin
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_session_begin, (dbh, username, password), 
      sqlo_db_handle_t dbh AND 
      const char * username AND
      const char * password )
{
  sqlo_db_ptr_t  dbp;
  int status;
  enum {MAX_BUFLEN = 1023};
  char uid[MAX_BUFLEN + 1];
  char pwd[MAX_BUFLEN + 1];
  char tnsname[MAX_BUFLEN + 1];

  if (dbh >= (int)_dbv_size || dbh < 0 || !_dbv[ dbh ]->used || !_dbv[ dbh ]->attached 
      || _dbv[ dbh ]->session_created ) {
    TRACE(1, fprintf(_trace_fp, "Invalid Database handle %d in sqlo_session_begin\n",
                     dbh);); 
    return SQLO_INVALID_DB_HANDLE;
  }

  dbp = _dbv[ dbh ];             /* setup the pointer to the db structure */
  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_session_begin[%d] starts\n", dbh); );

  if (!username || !password) {
    strcpy(dbp->errmsg, "sqlo_sesson_begin: No username or password specified");
    return SQLO_ERROR;
  }

  if (strchr(username, '/')) {
    if (SQLO_SUCCESS != sqlo_split_cstring(username, uid, pwd, tnsname, MAX_BUFLEN))
      return SQLO_ERROR;
  } else {
    strncpy(uid, username, (size_t)MAX_BUFLEN - 1);
    uid[MAX_BUFLEN] = '\0';
    strncpy(pwd, password, (size_t)MAX_BUFLEN - 1);
    pwd[MAX_BUFLEN] = '\0';
  }
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_session_begin: uid=%s, pwd=%s\n", uid, pwd););

  /* Set username attribute in session handle */
  status = OCIAttrSet((dvoid *)dbp->authp, OCI_HTYPE_SESSION,
                      (dvoid *) uid, (ub4)strlen(uid),
                      OCI_ATTR_USERNAME, dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_session_begin", "OCIAttrSet(username)");

  /* Set password attribute in user session handle */
  status = OCIAttrSet((dvoid *)dbp->authp, OCI_HTYPE_SESSION,
                      (dvoid *) pwd, (ub4)strlen(pwd),
                      OCI_ATTR_PASSWORD, dbp->errhp);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_session_begin", "OCIAttrSet(password)");

  /*
   * In non-blocking mode OCISessionBegin might return OCI_ERROR, but the
   * error code is 3123 (OCI_STILL_EXECUTING). 
   * We catch this case here and wait until we are connected
   */
  while ( OCI_STILL_EXECUTING == (status = OCISessionBegin((dvoid *) dbp->svchp, 
                                                           dbp->errhp, 
                                                           dbp->authp,
                                                           OCI_CRED_RDBMS, 
                                                           OCI_DEFAULT))
          || (status == OCI_ERROR && 
              _get_errcode(dbp) == ((-1) * OCI_STILL_EXECUTING) )) {
            TRACE(2, fprintf(_get_trace_fp(dbp),"sqlo_session_begin: "
                             "Still executing OCISessionBegin\n"););
            SQLO_USLEEP;
  }

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_session_begin", "OCISessionBegin");

  /* Set the authentication handle in the service context */
  while (OCI_STILL_EXECUTING == (status = OCIAttrSet((dvoid *)dbp->svchp, OCI_HTYPE_SVCCTX, 
                                                     (dvoid *)dbp->authp, (ub4) 0, OCI_ATTR_SESSION,
                                                      dbp->errhp)) ){
    TRACE(2, fprintf(_get_trace_fp(dbp),"Still executing OCIAttrSet (authp -> svchp)\n"););
    SQLO_USLEEP;

  }

  sqlo_geterrcode(dbh);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_session_begin", "OCIAttrSet(auth->service)");

  dbp->session_created = TRUE;

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_session_begin[%d]: logged in\n", dbh); );

  if (TRACE_ENABLED && _trace_level > 0)
    _open_trace_file(dbp);

  return (status); 
}


/*---------------------------------------------------------------------------
 *	sqlo_server_detach
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_server_detach, (dbh), sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;
  int status;

  if (dbh >= (int)_dbv_size || dbh < 0 || !_dbv[ dbh ]->used || !_dbv[ dbh ]->attached ) {
    TRACE(1, fprintf(_trace_fp, "Invalid Database handle %d in sqlo_server_detach\n",
                     dbh);); 
    return SQLO_INVALID_DB_HANDLE;
  }
  
  dbp = _dbv[dbh];             /* setup the pointer to the db structure */

  /* forgot to call sqlo_session_end ? */
  if (dbp->session_created) {
    status = sqlo_session_end(dbh);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_server_detach", "sqlo_session_end");
  }

  status = OCIServerDetach(dbp->srvhp, dbp->errhp, OCI_DEFAULT );

  if (OCI_SUCCESS != status) {
    CHECK_OCI_STATUS(dbp, status, "sqlo_server_detach", "OCIServerDetach");
    _db_release(dbp);
    return (status);
  }

  /* release the allocated resources */
  _db_release(dbp);

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_server_detach[%d]: detached\n", dbh); );

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 *	sqlo_session_end
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_session_end, (dbh),  sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;
  int status;
  
  if (dbh >= (int) _dbv_size || dbh < 0 || !_dbv[ dbh ]->used || 
      !_dbv[ dbh ]->session_created ) {
    TRACE(1, fprintf(_trace_fp, "Invalid Database handle %d in sqlo_session_end\n",
                     dbh);); 
    return SQLO_INVALID_DB_HANDLE;
  }

  
  dbp = _dbv[dbh];             /* setup the pointer to the db structure */

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_session_end[%d] starts\n", dbh); );


  /* close all open cursors  on this database connection */
  _close_all_db_cursors(dbp);

  while (OCI_STILL_EXECUTING == 
         (status = OCISessionEnd(dbp->svchp, dbp->errhp, dbp->authp, 0))) {
    TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_session_end: Still executing OCISessionEnd\n"););
    SQLO_USLEEP;
  }

  if (OCI_SUCCESS != status) {
    dbp->session_created = FALSE; /* mark it as finished */
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_session_end", "OCISessionEnd");
  }

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_session_end[%d] finished with success\n", dbh); );

  dbp->session_created = FALSE;

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 *	sqlo_connect
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_connect, (dbhp, cstr), 
      sqlo_db_handle_t * dbhp AND 
      const char * cstr)
{
  int status;
  enum {MAX_BUFLEN = 1023};
  char uid[MAX_BUFLEN + 1];
  char pwd[MAX_BUFLEN + 1];
  char tnsname[MAX_BUFLEN + 1];

  if (!dbhp)
    return(SQLO_ERROR);

  TRACE(2, fprintf(_get_trace_fp(NULL), "sqlo_connect starts\n"););

  if (SQLO_SUCCESS != sqlo_split_cstring(cstr, uid, pwd, tnsname, MAX_BUFLEN))
    return SQLO_ERROR;

  TRACE(3, fprintf(_get_trace_fp(NULL), "sqlo_connect: uid=%s, pwd=%s, tnsname=%s\n", 
                   uid, pwd, tnsname););

  if (SQLO_SUCCESS != (status = sqlo_server_attach(dbhp, tnsname)))
    return status;

  if (SQLO_SUCCESS != (status = sqlo_session_begin(*dbhp, uid, pwd))) {
    /* detach from server on error, otherwise the dbh is not freed */
    sqlo_server_detach(*dbhp);
    return status;
  }

  TRACE(2, fprintf(_get_trace_fp(_dbv[*dbhp]), "sqlo_connect[%d] finished with %d\n", *dbhp, status); );
  return (status); 
}


/*---------------------------------------------------------------------------
 *	sqlo_finish
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_finish, (dbh), sqlo_db_handle_t dbh)
{
  int status;
  sqlo_db_ptr_t  dbp;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_finish", SQLO_INVALID_DB_HANDLE);

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_finish[%d] starts\n", dbh); );

  if (dbp->session_created) {
    if (SQLO_SUCCESS == (status = sqlo_session_end(dbh))) {
      if (SQLO_SUCCESS != (status = sqlo_server_detach(dbh))) {
        TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_finish[%d] sqlo_server_detach failed\n", dbh););
        return status;
      }
    } else {
      int status2;

      if (SQLO_SUCCESS == (status2 = sqlo_server_detach(dbh))) {
        return status;          /* return the status of the session end failure */
      } else { 
        return status2;         /* return the status of the server detach failure */
      } /* end if SQLO_SUCCESS == sqlo_server_detach */
    } /* end if SQLO_SUCCESS == sqlo_session_end */
  } /* end if session created */

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_finish[%d] finished with success\n", dbh); );

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 *  sqlo_getdatabase
 *-------------------------------------------------------------------------*/
const char *
DEFUN(sqlo_getdatabase, (dbh) , sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_getdatabase", "Invalid dbh passed to sqlo_getdatabase");

  if (!dbp)
    return NULL;

  if (!dbp->tnsname) {
      sprintf(dbp->errmsg, "No tnsname in db structure\n");
      TRACE(2, (void) fputs(dbp->errmsg, _get_trace_fp(dbp)); );
  }

  return dbp->tnsname;
}


/*---------------------------------------------------------------------------
 *  sqlo_commit
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_commit, (dbh), sqlo_db_handle_t dbh )
{
  sqlo_db_ptr_t  dbp;
  int status;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_commit", SQLO_INVALID_DB_HANDLE);
  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_commit[%d]:\n", dbh); );
  
  status = OCITransCommit(dbp->svchp, dbp->errhp, OCI_DEFAULT);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sql_ocommit", "");

  return (status);
}


/*---------------------------------------------------------------------------
 *  sqlo_rollback
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_rollback, (dbh), sqlo_db_handle_t dbh )
{
  sqlo_db_ptr_t  dbp;
  int status;
  CHECK_DBHANDLE(dbp, dbh, "sqlo_rollback", SQLO_INVALID_DB_HANDLE);
  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_rollback[%d]:\n", dbh); );
  
  status = OCITransRollback(dbp->svchp, dbp->errhp, OCI_DEFAULT);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sql_orollback", "");

  return (SQLO_SUCCESS);
}


/*---------------------------------------------------------------------------
 *    sqlo_isopen
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_isopen, (sth), sqlo_stmt_handle_t sth)
{
  register  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_isopen", SQLO_INVALID_STMT_HANDLE);

  return (stp->opened ? SQLO_SUCCESS : 1 );
}    


/*---------------------------------------------------------------------------
 *	 sqlo_prepare 
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_prepare, (dbh, stmt), sqlo_db_handle_t dbh AND const char * stmt)
{
  sqlo_db_ptr_t dbp;
  sqlo_stmt_ptr_t stp;
  int status;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_prepare", SQLO_INVALID_DB_HANDLE);

  status = _stmt_new(dbp, stmt, &stp);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_prepare", "_stmt_new");

  TRACE(2, fprintf(_get_trace_fp(dbp), "Prepare: %s\n", stmt););

  /* prepare the statement and determine the statement type */
  status = _prepare(stp, stmt, &(stp->stype));

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_prepare", "_prepare");

  stp->prepared = TRUE;
  _bindpv_reset(stp);

  if (OCI_STMT_SELECT == stp->stype) {

    /* Set the prefetch count to _num_prefetch_rows */
    status = _set_prefetch_rows(stp, _num_prefetch_rows);
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_prepare", "_set_prefetch_rows");
  }

  return ((int) stp->sth);

}


/*---------------------------------------------------------------------------
 *	 sqlo_bind_by_name
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_bind_by_name, 
       (sth, param_name, param_type, param_addr, param_size, ind_addr, is_array),
       sqlo_stmt_handle_t sth AND const char * param_name AND int param_type AND 
       const void * param_addr AND unsigned int  param_size AND 
       short * ind_addr AND int is_array)
{
  register sqlo_stmt_ptr_t  stp;
  register OCIBind ** bindp_addr;
  register int status = SQLO_SUCCESS;

  CHECK_STHANDLE(stp, sth, "sqlo_bind_by_name", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, 
         fprintf(_get_trace_fp(stp->dbp), "sqlo_bind_by_name [%2d]: "
                  "name: %s type: %d (%s), size: %u\n", 
                  sth, param_name, param_type, _get_data_type_str(param_type), 
                  param_size););
  
  if (stp->prepared) {

    /* check for a refcursor */
    if (SQLOT_RSET == param_type) {
      status = sqlo_bind_ref_cursor(sth, param_name, (int *)param_addr);
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_by_name", "sqlo_bind_refcursor");
    } else {
      if (stp->num_bindpv >= stp->bindpv_size) {
        status = _alloc_bindp(stp, stp->num_bindpv+1);
        
        CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_by_name. alloc error for",
                         param_name);
      }
      bindp_addr = &stp->bindpv[ stp->num_bindpv ];
      
      status = OCIBindByName(stp->stmthp,
                             bindp_addr,
                             stp->dbp->errhp,
                             (text *) param_name,
                             (sb4) strlen(param_name),
                             (dvoid *) param_addr,
                             (sb4) param_size,
                             (ub2) param_type,
                             (dvoid *) ind_addr, 
                             (ub2 *) 0,
                             (ub2) 0,
                             (ub4) 0,
                             (ub4 *) 0,
                             OCI_DEFAULT);
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_by_name. Cannot bind", param_name);

      /* In case of arrays, we setup the skip parameters. */
      if (is_array) {
        status = OCIBindArrayOfStruct(*bindp_addr,
                                      stp->dbp->errhp, 
                                      param_size,
                                      ind_addr ? sizeof(short) : 0,
                                      0,
                                      0);
        CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_by_name. BindArrayOfStruct", param_name);
      }
      stp->num_bindpv++;
    }
  } else {
    status = SQLO_STMT_NOT_PARSED;
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_by_name", param_name);
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------
 *	 sqlo_bind_ref_cursor
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_bind_ref_cursor, (sth, cursor_name, sth2p),
      sqlo_stmt_handle_t sth AND 
      const char * cursor_name  AND
      sqlo_stmt_handle_t * sth2p
      )
{
  register sqlo_stmt_ptr_t  stp;
  sqlo_stmt_ptr_t st2p;
  register OCIBind ** bindp_addr;
  register int status = SQLO_SUCCESS;

  CHECK_STHANDLE(stp, sth, "sqlo_bind_ref_cursor", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, 
        fprintf(_get_trace_fp(stp->dbp), "sqlo_bind_ref_cursor [%2d]: name: %s\n", sth, cursor_name);)
  
  if (stp->prepared) {
    if (stp->num_bindpv >= stp->bindpv_size) {
      status = _alloc_bindp(stp, stp->num_bindpv+1);
      
      CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_ref_cursor. alloc error for",
                       cursor_name);
    }
    bindp_addr = &stp->bindpv[ stp->num_bindpv ];

    status = _stmt_new(stp->dbp, "", &st2p);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_ref_cursor", "_stmt_new");

    *sth2p = (int) st2p->sth;

    status = OCIHandleAlloc( (dvoid *) stp->dbp->envhp, (dvoid **) &st2p->stmthp,
                             OCI_HTYPE_STMT, (size_t) 0, (dvoid **) 0);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_ref_cursor", "OCIHandleAlloc(stmt)");


    status = OCIBindByName(stp->stmthp,
                           bindp_addr,
                           stp->dbp->errhp,
                           (text *) cursor_name,
                           strlen(cursor_name),
                           (dvoid *) &st2p->stmthp,
                           (sb4) 0,
                           (ub2) SQLT_RSET,
                           (dvoid *) 0, 
                           (ub2 *) 0,
                           (ub2 *) 0,
                           (ub4) 0,
                           (ub4 *) 0,
                           OCI_DEFAULT);

    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_bind_ref_cursor. Cannot bind", cursor_name);
    stp->num_bindpv++;

    st2p->cursor_type = REFCURSOR;
    st2p->prepared = TRUE;

  } else {
    CHECK_OCI_STATUS_RETURN(stp->dbp, SQLO_STMT_NOT_PARSED, "sqlo_bind_ref_cursor", cursor_name);
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------
 *	 sqlo_define_ntable
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_define_ntable, (sth, pos, sth2p),
      sqlo_stmt_handle_t sth AND 
      unsigned int pos AND
      sqlo_stmt_handle_t * sth2p
      )
{
  register sqlo_stmt_ptr_t  stp;
  sqlo_stmt_ptr_t st2p;
  register int status = SQLO_SUCCESS;

  CHECK_STHANDLE(stp, sth, "sqlo_define_ntable", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, fprintf(_get_trace_fp(stp->dbp), 
                   "sqlo_define_ntable [%2d]: pos: %u\n", sth, pos););
  
  if (stp->prepared) {
    status = _alloc_definep(stp, pos);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_define_ntable. alloc error", "");

    status = _stmt_new(stp->dbp, "", &st2p);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_define_ntable", "_stmt_new");
  
    *sth2p = (int) st2p->sth;
  
    status = OCIHandleAlloc( (dvoid *) stp->dbp->envhp, (dvoid **) &st2p->stmthp,
                             OCI_HTYPE_STMT, (size_t) 0, (dvoid **) 0);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_define_ntable", "OCIHandleAlloc(stmt)");
  
  
    status = OCIDefineByPos(stp->stmthp,
                            &stp->defnpv[ pos - 1 ],
                            stp->dbp->errhp,
                            (ub4) pos,
                            (dvoid *) &st2p->stmthp,
                            (sword) 0,
                            (ub2)SQLT_RSET,
                            (dvoid *) 0, 
                            (ub2 *) 0,
                            (ub2 *) 0,
                            OCI_DEFAULT);
  
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_define_ntable: Cannot define", "");
    stp->num_defnpv++;

    st2p->cursor_type = NTABLE;
    st2p->prepared = TRUE;
  
  } else {
    CHECK_OCI_STATUS_RETURN(stp->dbp, SQLO_STMT_NOT_PARSED, "sqlo_define_ntable", "");
  }
  return (status == OCI_SUCCESS ? SQLO_SUCCESS : status);
}


/*---------------------------------------------------------------------------
 *	 sqlo_bind_by_pos
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_bind_by_pos, 
       (sth, param_pos, param_type, param_addr, param_size, ind_addr, is_array),
       sqlo_stmt_handle_t sth AND 
       int param_pos AND 
       int param_type AND 
       const void * param_addr AND 
       unsigned int  param_size AND 
       short * ind_addr AND 
       int is_array)
{
  register sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_bind_by_pos", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, 
      fprintf(_get_trace_fp(stp->dbp), 
              "sqlo_bind_by_pos [%2d]: pos: %d type: %d (%s), size: %u\n", 
               sth, param_pos, param_type, _get_data_type_str(param_type), param_size););
  
  return(_bind_by_pos(stp, (unsigned int)param_pos, param_type, param_addr, param_size, 
                       ind_addr, is_array));

}


/*---------------------------------------------------------------------------
 *	 sqlo_bind_by_pos2
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_bind_by_pos2, 
      (sth, param_pos, param_type, param_addr, param_size, ind_addr, rcode_addr, 
       skip_size),
       sqlo_stmt_handle_t sth AND 
       int param_pos AND 
       int param_type AND 
       const void * param_addr AND 
       unsigned int  param_size AND 
       short * ind_addr AND 
       unsigned short * rcode_addr AND
       unsigned int skip_size)
{
  register sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_bind_by_pos2", SQLO_INVALID_STMT_HANDLE);

  TRACE(3,
        fprintf(_get_trace_fp(stp->dbp), "sqlo_bind_by_pos2 [%2d]: "
                "pos: %d type: %d (%s), size: %u, skip_size: %u\n", 
                sth, param_pos, param_type, _get_data_type_str(param_type), 
                param_size, skip_size););
  
  return(_bind_by_pos2(stp, (unsigned int)param_pos, param_type, param_addr, param_size, 
                       ind_addr, rcode_addr, skip_size));

}


/*---------------------------------------------------------------------------
 *	 sqlo_define_by_pos
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_define_by_pos, 
       (sth, value_pos, value_type, value_addr, value_size, ind_addr, rlen_addr,
        is_array),
       sqlo_stmt_handle_t sth AND
       int value_pos AND 
       int value_type AND 
       const void * value_addr 
       AND unsigned int  value_size AND 
       short * ind_addr AND
       short * rlen_addr AND /* where the returned length is stored. */
       int is_array)
{
  register sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_define_by_pos", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, 
      fprintf(_get_trace_fp(stp->dbp), 
              "sqlo_define_by_pos [%2d]: pos: %d type: %d (%s), "
              "size: %u, is_array: %d\n", 
              sth, value_pos, value_type, _get_data_type_str(value_type), 
              value_size, is_array););

  /* check for a nested table */
  if (SQLOT_RSET == value_type ) {
    return(sqlo_define_ntable(sth, (unsigned int)value_pos, (int *) value_addr));
  } else {
    return(_define_by_pos(stp, (unsigned int)value_pos, value_type, value_addr, 
                           value_size, ind_addr, (ub2*)rlen_addr, (ub2*) 0, is_array));
  }
}


/*---------------------------------------------------------------------------
 *	 sqlo_define_by_pos2
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_define_by_pos2, 
       (sth, value_pos, value_type, value_addr, value_size, ind_addr, rlen_addr,
        rcode_addr, skip_size),
       sqlo_stmt_handle_t sth AND
       int value_pos AND 
       int value_type AND 
       const void * value_addr  AND
       unsigned int  value_size AND 
       short * ind_addr AND
       unsigned short * rlen_addr AND 
       unsigned short * rcode_addr AND
       unsigned int skip_size
       )
{
  register sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_define_by_pos2", SQLO_INVALID_STMT_HANDLE);

  TRACE(3, 
         fprintf(_get_trace_fp(stp->dbp), "sqlo_define_by_pos [%2d]: pos: %d type: %d (%s), "
                 "size: %u, skip_size: %u\n", 
                 sth, value_pos, value_type, _get_data_type_str(value_type), 
                 value_size, skip_size););
  
  return(_define_by_pos2(stp, (unsigned int) value_pos, value_type, value_addr, 
                          value_size, ind_addr, (ub2*)rlen_addr, (ub2*) rcode_addr, skip_size));

}


/*---------------------------------------------------------------------------
 * sqlo_execute
 *--------------------------------------------------------------------------*/
int 
DEFUN(sqlo_execute, (sth, iterations), 
      sqlo_stmt_handle_t sth AND 
      unsigned int iterations )
{
  register sqlo_stmt_ptr_t  stp;
  int status = SQLO_SUCCESS;

  CHECK_STHANDLE(stp, sth, "sqlo_execute", SQLO_INVALID_STMT_HANDLE);

  EXEC_WHEN_THREADING
    (
     if (TRUE != _thread_id_equal( stp->thread_id, _get_thread_id() ) ) {
       sprintf(stp->dbp->errmsg,
               "sqlo_execute [%2d]: Try to execute a cursor opened by another thread !!!\n", 
               sth);
       TRACE(1, fprintf(_get_trace_fp(stp->dbp), stp->dbp->errmsg););
       return SQLO_INVALID_STMT_HANDLE;
     }
     );


  TRACE(3, fprintf(_get_trace_fp(stp->dbp), "sqlo_execute [%2d]: iter=%u, stmt=%.80s\n", 
                   sth, iterations, _get_stmt_string(stp)););

  /* For REF CURSORS and NESTED TABLES, we determine the statement type and
   * define the output here
   */
  if (DEFAULT != stp->cursor_type && 0 == stp->num_executions ) {
    /* REF CURSOR or NESTED TABLE */
    status = OCIAttrGet((dvoid*) stp->stmthp, (ub4) OCI_HTYPE_STMT, 
                        (dvoid*) &(stp->stype),(ub4 *) 0, 
                        (ub4) OCI_ATTR_STMT_TYPE, 
                        (OCIError *) stp->dbp->errhp );

    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_execute", "GetStmtType");
    status = _define_output(stp);
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_execute", "_define_output");

  } else  if (stp->prepared ) {
    /* standard cursor */
    status = OCIStmtExecute(stp->dbp->svchp, 
                            stp->stmthp, 
                            stp->dbp->errhp, 
                            (ub4) iterations, 
                            (ub4) 0,
                            (OCISnapshot *) 0, 
                            (OCISnapshot *) 0,
                            (ub4) OCI_DEFAULT);
      
    if (OCI_SUCCESS != status && OCI_NO_DATA != status) {
      if (OCI_STILL_EXECUTING == status) {
        stp->still_executing = TRUE;
        stp->opened          = TRUE;
        status               = SQLO_STILL_EXECUTING;
      }

      stp->dbp->status = status;
      _save_oci_status(stp->dbp, "sqlo_execute", _get_stmt_string(stp), __LINE__);
    } else {
      stp->opened          = TRUE;
      stp->still_executing = FALSE;
      stp->dbp->status     = status;
      _bindpv_reset(stp);

    } /* end if OCI_SUCCESS != status  */
  } else {
    CHECK_OCI_STATUS_RETURN(stp->dbp, SQLO_STMT_NOT_PARSED, "sqlo_execute", "");
  }
  TRACE(3, fprintf(_get_trace_fp(stp->dbp), "sqlo_execute returns %d\n", status););
  return (status);
}


/*---------------------------------------------------------------------------
 * sqlo_ocol_names
 *-------------------------------------------------------------------------*/
const char ** 
DEFUN(sqlo_ocol_names, (sth, num),
      sqlo_stmt_handle_t sth AND 
      int * num)
{

  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_ocol_names", NULL);

  if (stp->stype != OCI_STMT_SELECT || !stp->opened)  {
    sprintf(stp->dbp->errmsg, 
            "Cannot get values for a non-select/non-opened statement (sth %d) passed to sqlo_ocol_names stype=%u opened=%d\n", sth, (unsigned int) stp->stype, stp->opened);
    TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
    if (num)
      *num = 0;
	return NULL;
  }

  if (num)
    *num = (int) stp->num_defnpv;

  TRACE(2, fprintf(_get_trace_fp(stp->dbp), 
                   "sqlo_ocol_names: Returning %u column names\n",
                   stp->num_defnpv););

  return ((const char **)stp->ocol_namev);
}


/*---------------------------------------------------------------------------
 *	sqlo_ocol_name_lens 
 *-------------------------------------------------------------------------*/
const int *
DEFUN(sqlo_ocol_name_lens, (sth, num),
      sqlo_stmt_handle_t sth AND 
      int * num)
{

  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_ocol_name_lens", NULL);

  if (stp->stype != OCI_STMT_SELECT || !stp->opened) {
      sprintf(stp->dbp->errmsg, 
              "Cannot get values for a non-select/non-opened statement (sth %d) passed to sqlo_ocol_name_lens\n", sth);
      TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
	return NULL;
  }

  if (num)
    *num = (int)stp->num_defnpv;
  
  TRACE(2, fprintf(_get_trace_fp(stp->dbp), 
                   "sqlo_ocol_name_lens: Returning %u column names\n",
                   stp->num_defnpv););

  return ((const int *)stp->ocol_namev_size);
}


/*---------------------------------------------------------------------------
 *	sqlo_value_lens
 *-------------------------------------------------------------------------*/
const unsigned short *
DEFUN(sqlo_value_lens, (sth, num),
      sqlo_stmt_handle_t sth AND 
      int * num)
{

  sqlo_stmt_ptr_t  stp;

  CHECK_STHANDLE(stp, sth, "sqlo_value_lens", NULL);

  if (stp->stype != OCI_STMT_SELECT || !stp->opened) {
    sprintf(stp->dbp->errmsg, 
            "Cannot get value len for a non-select/non-opened statement (sth %d) passed to sqlo_value_lens\n", sth);
    TRACE(1, (void) fputs(stp->dbp->errmsg, _get_trace_fp(stp->dbp)););
    if (num)
      *num = 0;
	return NULL;
  }

  if (num)
    *num = (int)stp->num_defnpv;
  
  TRACE(2, fprintf(_get_trace_fp(stp->dbp), "sqlo_value_lens: Returning %u items\n",
                     stp->num_defnpv););

  return ((const unsigned short *) stp->rlenv);
}


/*---------------------------------------------------------------------------
 *	sqlo_get_oci_handle 
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_get_oci_handle, (sqloh, ocihp, type),
      int sqloh AND void * ocihp AND 
      sqlo_oci_handle_types_e type)
{
  sqlo_db_ptr_t   dbp;
  sqlo_stmt_ptr_t stp;

  switch (type) {
    /* global handles */
  case SQLO_OCI_HTYPE_ENV:
    CHECK_DBHANDLE(dbp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_DB_HANDLE);
    *((OCIEnv **) ocihp) = dbp->envhp;
    break;

    /* Connection specific handles */
  case SQLO_OCI_HTYPE_ERROR:
    CHECK_DBHANDLE(dbp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_DB_HANDLE);
    *((OCIError **) ocihp) = dbp->errhp;
    break;

  case SQLO_OCI_HTYPE_SVCCTX:
    CHECK_DBHANDLE(dbp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_DB_HANDLE);
    * ((OCISvcCtx **) ocihp) = dbp->svchp;
    break;

  case SQLO_OCI_HTYPE_SERVER:
    CHECK_DBHANDLE(dbp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_DB_HANDLE);
    * ((OCIServer **) ocihp) = dbp->srvhp;
    break;

  case SQLO_OCI_HTYPE_SESSION:
    CHECK_DBHANDLE(dbp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_DB_HANDLE);
    *(( OCISession **) ocihp) = dbp->authp;

    break;

    /* Statement specific handles */
  case SQLO_OCI_HTYPE_STMT:
    CHECK_STHANDLE(stp, sqloh, "sqlo_get_oci_handle", SQLO_INVALID_STMT_HANDLE);
    * ((OCIStmt **) ocihp) = stp->stmthp;
    break;

  default:
    return (SQLO_INVALID_OCI_HANDLE_TYPE);
    break; /* to keep the compiler happy */
  }
  return (SQLO_SUCCESS);
}


/*---------------------------------------------------------------------------
 * sqlo_get_db_handle
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_get_db_handle, (sth), sqlo_stmt_handle_t sth)
{
  sqlo_stmt_ptr_t stp;
  CHECK_STHANDLE(stp, sth, "sqlo_get_db_handle", SQLO_INVALID_STMT_HANDLE);

  if (!stp->used || NULL == stp->dbp) {
    sprintf(_errmsg, "Invalid sth %d passed to sqlo_get_db_handle.\n", sth);
    TRACE(1, (void) fputs(_errmsg, _trace_fp););
    return (SQLO_INVALID_STMT_HANDLE);
  }
  return( (int) stp->dbp->dbh);
}


/*---------------------------------------------------------------------------
 *	sqlo_version - checks if the version is sufficient
 *
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_version, (version_str), const char * version_str)
{
  unsigned int major, minor, micro;
  enum {TMP_VERSION_LEN = 64};
  char tmp_version[TMP_VERSION_LEN+1];

  /* HP/UX 9 (%@#!) writes to sscanf strings */
  strncpy(tmp_version, version_str, TMP_VERSION_LEN);
  tmp_version[TMP_VERSION_LEN] = '\0';

  if (sscanf(tmp_version, "%u.%u.%u", &major, &minor, &micro) != 3) {
    return SQLO_MALFORMED_VERSION_STR;
   }

  if ( sqlo_major_version ==  major &&
       sqlo_minor_version == minor &&
       sqlo_micro_version >= micro) {
      return SQLO_SUCCESS;
    }
  return SQLO_WRONG_VERSION;
}


/*---------------------------------------------------------------------------
 *	sqlo_set_blocking
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_set_blocking, (dbh, on), sqlo_db_handle_t dbh AND unsigned int on)
{
  sqlo_db_ptr_t dbp;
  int status;
  unsigned int new_mode;
  unsigned int blocking;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_set_blocking", SQLO_INVALID_DB_HANDLE);


  TRACE(3, fprintf(_get_trace_fp(dbp),"sqlo_set_blocking: on: %u\n", on););
  status = _get_blocking_mode(dbp, &blocking);

  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_set_blocking:", "_get_blocking_mode[1]");

  new_mode = on > 0 ? SQLO_ON : SQLO_OFF;

  if ( blocking != new_mode)
    {
      /* toggle the mode */
      if (OCI_SUCCESS != 
          (status = OCIAttrSet((dvoid *) dbp->srvhp, (ub4) OCI_HTYPE_SERVER, 
                               (dvoid *) 0, (ub4) 0, 
                               (ub4) OCI_ATTR_NONBLOCKING_MODE,
                               dbp->errhp))) { 
        TRACE(2, fprintf(_get_trace_fp(dbp), "Unable to toggle blocking mode"););
        CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_set_blocking", "");
        return (SQLO_ERROR);
      } 
      
    }

  status = _get_blocking_mode(dbp, &blocking);

  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_set_blocking:", "_get_blocking_mode[2]");
    
  TRACE(3, fprintf(_get_trace_fp(dbp),"sqlo_set_blockin_oci_init_mode: mode is %s\n",
                   ( SQLO_ON == blocking ? "blocking" : "non-blocking")););

  return (SQLO_SUCCESS);
  
}


/*---------------------------------------------------------------------------
 * sqlo_get_blocking
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_get_blocking, (dbh, blocking), 
      sqlo_db_handle_t dbh AND unsigned int *blocking )
{
  sqlo_db_ptr_t  dbp;
  int status;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_get_blocking", SQLO_INVALID_DB_HANDLE);

  status = _get_blocking_mode(dbp, blocking);

  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_get_blocking", "OCIAttrGet");

  return (status);
}


/*---------------------------------------------------------------------------
 *	sqlo_break
 *-------------------------------------------------------------------------*/
int 
DEFUN(sqlo_break, (dbh), sqlo_db_handle_t dbh)
{
  sqlo_db_ptr_t  dbp;
  int status;
  unsigned int blocking;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_break", SQLO_INVALID_DB_HANDLE);

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_break\n"););

  status = _get_blocking_mode(dbp, &blocking);
  if (status < 0)
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_break", "_get_blocking_mode");
    
  if ( SQLO_OFF == blocking ) {
    status = OCIBreak(dbp->srvhp, dbp->errhp);
    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_break OCIBreak returned %d\n", status););

    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_break", "OCIBreak");
    
    status = OCIReset(dbp->srvhp, dbp->errhp);
    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_break OCIReset returned %d\n", status););
             

    /*      CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_break", "OCIReset");  */

    EXEC_WHEN_THREADING(_stmtv_lock();); /* start of critical section */
      
    /* close all open cursors  on this database connection */
    _close_all_executing_cursors(dbp);

    /* Did sqlo_exec open a cursor */
    if (dbp->stmthp) {
      status = OCIHandleFree(dbp->stmthp, OCI_HTYPE_STMT); 
      dbp->stmthp = 0;
      CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_break", "OCIHandleFree(dbp->stmthp)");
    }

    EXEC_WHEN_THREADING(_stmtv_unlock();); /* start of critical section */

  } else {
    status = SQLO_SUCCESS;
  }

  TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_break  returns %d\n", status););
  return(status);

}


/*---------------------------------------------------------------------------
 * sqlo_alloc_lob_desc
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_alloc_lob_desc, (dbh, loblpp), 
      sqlo_db_handle_t dbh     AND           /* I - The database handle */
      sqlo_lob_desc_t *loblpp   /* O - The allocated lob locator */
      )
{
  int status;
  sqlo_db_ptr_t  dbp;
  
  CHECK_DBHANDLE(dbp, dbh, "sqlo_alloc_lob_desc", SQLO_INVALID_DB_HANDLE);

  status = OCIDescriptorAlloc((dvoid *) dbp->envhp, (dvoid **) loblpp,
                         (ub4) OCI_DTYPE_LOB,
                         (size_t) 0, (dvoid **) 0);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_alloc_lob_desc", "OCIDescriptorAlloc");

  return (SQLO_SUCCESS);
} 


/*---------------------------------------------------------------------------
 * sqlo_free_lob_desc
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_free_lob_desc, (dbh, loblpp), 
      sqlo_db_handle_t dbh     AND           /* I - The database handle */
      sqlo_lob_desc_t *loblpp   /* I - The lob locator */
      )
{
  int status;
  sqlo_db_ptr_t  dbp;
  
  CHECK_DBHANDLE(dbp, dbh, "sqlo_free_lob_desc", SQLO_INVALID_DB_HANDLE);

  if (!loblpp || !*loblpp) {
    strcpy(dbp->errmsg, "Invalid lob locator passed to sqlo_free_lob_desc");
    return SQLO_ERROR;
  }

  status = OCIDescriptorFree((dvoid **) *loblpp, (ub4) OCI_DTYPE_LOB);

  *loblpp = NULL;
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_free_lob_desc", "OCIDescriptorAlloc");
  
  return (SQLO_SUCCESS);
} 


/*---------------------------------------------------------------------------
 * sqlo_lob_write_buffer
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_lob_write_buffer, (dbh, loblp, loblen, bufp, bufl, piece), 
      sqlo_db_handle_t dbh       AND         /* I - The database handle */
      sqlo_lob_desc_t loblp   AND /* I - The allocated lob locator */
      unsigned int loblen AND   /* I - The size of the LOB */
      void *bufp    AND         /* I - The lob data to write into the lob */
      unsigned int bufl AND     /* I - The number of bytes in the buffer  */
      unsigned int piece        /* I - SQLO_ONE_PIECE, SQLO_FIRST_PIECE, SQLO_NEXT_PIECE or SQLO_LAST_PIECE */
      )
{
  int status;
  sqlo_db_ptr_t  dbp;
  ub4 amtp = loblen;
  ub4 offset = 1;
  ub4 nbytes;
  ub1 p;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_write_buffer", SQLO_INVALID_DB_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_buffer: loblen=%u, len=%u, %s piece\n", loblen,
                   bufl,
                   (piece == SQLO_ONE_PIECE) ? "ONE" : 
                   (piece == SQLO_FIRST_PIECE) ? "FIRST" : 
                   (piece == SQLO_NEXT_PIECE) ? "NEXT" : 
                   (piece == SQLO_LAST_PIECE) ? "LAST" : "???"
                    ););

  nbytes = (loblen < bufl ? loblen : bufl);

  if (loblen <= bufl &&  SQLO_FIRST_PIECE == piece ) 
    p = OCI_ONE_PIECE;
  else
    p = (ub1) piece;

  status = OCILobWrite(dbp->svchp,
                       dbp->errhp,
                       loblp,
                       &amtp,
                       offset,
                       (dvoid *) bufp,
                       (ub4) nbytes,
                       p,
                       (dvoid *) 0,
                       (OCICallbackLobWrite) NULL,
                       (ub2) 0,
                       (ub1) SQLCS_IMPLICIT);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_write_buffer", "OCILobWrite");

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 * sqlo_lob_append_buffer
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_lob_append_buffer, (dbh, loblp, loblen, bufp, bufl, piece), 
      sqlo_db_handle_t dbh       AND         /* I - The database handle */
      sqlo_lob_desc_t loblp   AND /* I - The allocated lob locator */
      unsigned int loblen AND   /* I - The size of the LOB */
      void *bufp    AND         /* I - The lob data to write into the lob */
      unsigned int bufl AND     /* I - The number of bytes in the buffer  */
      unsigned int piece        /* I - SQLO_ONE_PIECE, SQLO_FIRST_PIECE, SQLO_NEXT_PIECE or SQLO_LAST_PIECE */
      )
#ifdef HAVE_OCILOBWRITEAPPEND
{
  int status;
  sqlo_db_ptr_t  dbp;
  ub4 amtp = loblen;
  ub4 nbytes;
  ub1 p;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_append_buffer", SQLO_INVALID_DB_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_append_buffer: loblen=%u, len=%u, %s piece\n", loblen,
                   bufl,
                   (piece == SQLO_ONE_PIECE) ? "ONE" : 
                   (piece == SQLO_FIRST_PIECE) ? "FIRST" : 
                   (piece == SQLO_NEXT_PIECE) ? "NEXT" : 
                   (piece == SQLO_LAST_PIECE) ? "LAST" : "???"
                    ););

  nbytes = (loblen < bufl ? loblen : bufl);

  if (loblen <= bufl &&  SQLO_FIRST_PIECE == piece ) 
    p = OCI_ONE_PIECE;
  else
    p = (ub1) piece;

  status = OCILobWriteAppend(dbp->svchp,
                             dbp->errhp,
                             loblp,
                             &amtp,
                             (dvoid *) bufp,
                             (ub4) nbytes,
                             p,
                             (dvoid *) 0,
                             (OCICallbackLobWrite) NULL,
                             (ub2) 0,
                             (ub1) SQLCS_IMPLICIT);

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_append_buffer", "OCILobWriteAppend");

  return SQLO_SUCCESS;
}
#else
{
  sqlo_db_ptr_t  dbp;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_append_buffer", SQLO_INVALID_DB_HANDLE);

  /* OCILobWriteAppend is not available. We return an error in this case. */
  CHECK_OCI_STATUS_RETURN(dbp, OCI_ERROR, "sqlo_lob_append_buffer", "OCILobWriteAppend not available in this Oracle version");
}
#endif

/*---------------------------------------------------------------------------
 * sqlo_lob_write_stream
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_lob_write_stream, (dbh, loblp, filelen, fp), 
      sqlo_db_handle_t dbh       AND         /* I - The database handle */
      sqlo_lob_desc_t loblp   AND /* I - The allocated lob locator */
      unsigned int filelen  AND /* I - The size of the file */
      FILE *fp                  /* I - The filepointer */
      )
{
  int status;
  sqlo_db_ptr_t  dbp;
  unsigned char buf[MAX_LONG_SIZE]; /* The buffer */
  unsigned int nbytes;          /* number of bytes to read from stream */
  unsigned int remainder = filelen; /* The number of bytes left */
  unsigned int piece;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_write_stream", SQLO_INVALID_DB_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_stream: filelen=%u\n", filelen););

  if (filelen > MAX_LONG_SIZE)
    nbytes = MAX_LONG_SIZE;
  else
    nbytes = filelen;

  /* get a chunk of data */
  if (fread( (void*) buf, (size_t)nbytes, 1, fp) != 1 ) {
    strcpy(dbp->errmsg, "sqlo_lob_write_stream: I/O error. Could not get data from stream");
    return SQLO_ERROR;
  }

  remainder -= nbytes;


  if (0 == remainder) {           /* excatly one piece in the file */
    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_stream: nbytes=%u remain=%u ONE piece\n",
                     nbytes, remainder
                     ););
    status = SQLO_SUCCESS;
    piece = SQLO_ONE_PIECE;
    do {

      if (SQLO_STILL_EXECUTING == status) {
        SQLO_USLEEP;
        ;
      }

      status = sqlo_lob_write_buffer(dbh, loblp, nbytes, (void *)buf, nbytes, piece);
    } while (SQLO_STILL_EXECUTING == status);

    return status;

  } else {
    
#ifdef HAVE_OCILOBOPEN
    /* wrap this by an OCILobOpen()/OCILobClose() pair to make sure
     * the indices are uptdated only once
     */
    status = OCILobOpen(dbp->svchp, dbp->errhp, loblp, OCI_LOB_READWRITE);

    if (0 > status) {
      if (_get_errcode(dbp) != ((-1) * SQLO_STILL_EXECUTING) )
        CHECK_OCI_STATUS_RETURN(dbp, status, "slqo_lob_write_stream", "OCILobOpen");
    }
#endif

    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_stream: "
                     "nbytes=%u remain=%u FIRST piece\n",
                     nbytes, remainder
                     ););
    /* more than one piece, insert first */

    piece = SQLO_FIRST_PIECE;
    /* wait until non-blocking call finished */
    do {

      if (SQLO_STILL_EXECUTING == status) {
        SQLO_USLEEP;
        ;
      }

      status = sqlo_lob_write_buffer(dbh, loblp, filelen, (void *)buf, nbytes, piece);
      
    } while (SQLO_STILL_EXECUTING == status);

    if (SQLO_NEED_DATA != status) {
      if (0 > status) {
        CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_write_stream", "sqlo_lob_write_buffer(FIRST)");
      } else {
        sprintf(dbp->errmsg,"sqlo_lob_write_buffer returned %d, expected %d (SQLO_NEED_DATA)",
                status, SQLO_NEED_DATA);
        return SQLO_ERROR;
      }
    }

    /* insert remaining pieces */
    piece = SQLO_NEXT_PIECE;
    do {
      if (remainder > MAX_LONG_SIZE)
        nbytes = MAX_LONG_SIZE;
      else {
        nbytes = remainder;
        piece = SQLO_LAST_PIECE;
      }

      if (fread( (void*) buf, (size_t)nbytes, 1, fp) != 1 ) {
        strcpy(dbp->errmsg, "sqlo_lob_write_stream: I/O error. Could not get data from stream");
        TRACE(2, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_stream: "
                         "Error during fread(). Setting piece to SQLO_LAST_PIECE\n"););

        piece = SQLO_LAST_PIECE;
      }

      TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_write_stream: "
                       "nbytes=%u remain=%u, %s piece\n",
                       nbytes, remainder - nbytes,
                       (piece == SQLO_ONE_PIECE) ? "ONE" : 
                       (piece == SQLO_FIRST_PIECE) ? "FIRST" : 
                       (piece == SQLO_NEXT_PIECE) ? "NEXT" : 
                       (piece == SQLO_LAST_PIECE) ? "LAST" : "???"
                       ););
      /* wait until non-blocking call finished */
      do {
        
        if (SQLO_STILL_EXECUTING == status) {
          SQLO_USLEEP;
          ;
        }
        status = sqlo_lob_write_buffer(dbh, loblp, filelen, (void *)buf, nbytes, piece);

      } while (SQLO_STILL_EXECUTING == status);

      if ( 0 > status) {
        CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_write_stream", "sqlo_lob_write_buffer(NEXT/LAST)");
      }

      remainder -= nbytes;
    } while (status == SQLO_NEED_DATA && ! feof(fp));

#ifdef HAVE_OCILOBOPEN
    while (SQLO_STILL_EXECUTING == 
           (status = OCILobClose(dbp->svchp, dbp->errhp, loblp))) {
      SQLO_USLEEP;
    }
    CHECK_OCI_STATUS_RETURN(dbp, status, "slqo_lob_write_stream", "OCILobClose");
#endif
  } /* endif remainder == 0 */

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_write_stream", "OCILobWrite");

  return SQLO_SUCCESS;
}


/*---------------------------------------------------------------------------
 * sqlo_lob_get_length
 *-------------------------------------------------------------------------*/
int                             /* O - status */
DEFUN(sqlo_lob_get_length, (dbh, loblp, loblenp),
      sqlo_db_handle_t dbh               AND /* I - database handle */
      sqlo_lob_desc_t loblp AND /* I - Lob locator */
      unsigned int * loblenp    /* O - The length of the lob */
      )
{
  sqlo_db_ptr_t dbp;
  int status;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_get_length", SQLO_INVALID_DB_HANDLE);

  if (!loblenp) {
    strcpy(dbp->errmsg, "sqlo_lob_get_length: NULL pointer passed in loblen");
    return SQLO_ERROR;
  }

  status = OCILobGetLength(dbp->svchp,
                           dbp->errhp,
                           (OCILobLocator *) loblp,
                           (ub4*) loblenp);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_get_length", "OCILobGetLength");
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_get_length: loblen=%u\n", *loblenp););
  return status;
}


/*---------------------------------------------------------------------------
 * sqlo_lob_read_buffer
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS, SQLO_NEED_DATA or < 0 0n error */
DEFUN(sqlo_lob_read_buffer, (dbh, loblp, loblen, bufp, bufl), 
      sqlo_db_handle_t dbh       AND         /* I - The database handle */
      sqlo_lob_desc_t loblp   AND /* I - The allocated lob locator */
      unsigned int loblen AND   /* I - The max len of the lob  */
      void *bufp    AND         /* I - The lob data to write into the lob */
      unsigned int bufl         /* I - The capacity of buf  */
      )
{
  int status;
  sqlo_db_ptr_t dbp;
  ub4 amtp = 0;
  ub4 offset = 1;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_read_buffer", SQLO_INVALID_DB_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_buffer: loblen=%u, buflen=%u\n", loblen,bufl););

  
  amtp = loblen;

  status = OCILobRead(dbp->svchp,
                      dbp->errhp,
                      loblp,
                      &amtp,
                      offset,
                      (dvoid *) bufp,
                      (ub4) (loblen < bufl ? loblen : bufl),
                      (dvoid *) 0,
                      (OCICallbackLobRead) NULL,
                      (ub2) 0,
                      (ub1) SQLCS_IMPLICIT);

  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_buffer: amtp=%u\n", amtp););

  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_read", "OCILobRead");

  return status;
}


/*---------------------------------------------------------------------------
 * sqlo_lob_read_stream
 *-------------------------------------------------------------------------*/
int                             /* O - Status SQLO_SUCCESS or SQLO_INVALID_DB_HANDLE */
DEFUN(sqlo_lob_read_stream, (dbh, loblp, loblen, fp), 
      sqlo_db_handle_t dbh       AND         /* I - The database handle */
      sqlo_lob_desc_t loblp   AND /* I - The allocated lob locator */
      unsigned int loblen  AND  /* I - The size of the lob */
      FILE *fp                  /* I - The filepointer, where we write the data. */
      )
{
  int status = SQLO_SUCCESS;
  sqlo_db_ptr_t dbp;
  unsigned char buf[MAX_LONG_SIZE]; /* The buffer */
  unsigned int nbytes;              /* number of bytes to read from stream */
  unsigned int remainder;       /* The number of bytes left */

  CHECK_DBHANDLE(dbp, dbh, "sqlo_lob_read_stream", SQLO_INVALID_DB_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_stream: loblen=%u\n", loblen););

  if (loblen > MAX_LONG_SIZE)
    nbytes = MAX_LONG_SIZE;
  else
    nbytes = loblen;

  /* get a chunk of data out of the lob */
  do {
    if (SQLO_STILL_EXECUTING == status) {
      SQLO_USLEEP;
      ;
    }
    status = sqlo_lob_read_buffer(dbh, loblp, loblen, (void *)buf, nbytes);

  } while (SQLO_STILL_EXECUTING == status);

  switch (status) {
  case SQLO_SUCCESS:
    /* got all in one piece */
    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_stream: got all in one piece (%u bytes)\n", loblen););
    (void) fwrite( (void *) buf, (size_t)loblen, 1, fp);
    break;
  case SQLO_ERROR:
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_read_stream", "sqlo_lob_read_buffer(FIRST)");
    return status;
    break;
  case SQLO_NEED_DATA:
    
    remainder = loblen;
    TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_stream: got first piece (%u bytes)\n", nbytes););
    /* write this buffer */
    (void) fwrite( (void *) buf, (size_t)nbytes, 1, fp);
    
    do {
      memset(buf, '\0', MAX_LONG_SIZE);
      /*      loblen = 0;*/

      remainder -= nbytes;

      /* get a chunk of data out of the lob */
      do {
        if (SQLO_STILL_EXECUTING == status) {
          SQLO_USLEEP;
          ;
        }
    
        status = sqlo_lob_read_buffer(dbh, loblp, loblen, (void *)buf, nbytes);

      } while (SQLO_STILL_EXECUTING == status);

      /* write the data to the file */
      if (remainder < nbytes) {
        TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_stream: "
                         "got last piece (%u bytes)\n", remainder););

        (void) fwrite( (void *) buf, (size_t)remainder, 1, fp);
      } else {

        TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_lob_read_stream: "
                         "got next piece (%u bytes)\n", nbytes););
        (void) fwrite( (void *) buf, (size_t)nbytes, 1, fp);
      }

    } while (SQLO_NEED_DATA == status);
    break;

  default:
    CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_lob_read_stream", "");
    break;

  } 
  return status;
}


/*-------------------------------------------------------------------------
 * sqlo_set_prefetch_rows
 *-----------------------------------------------------------------------*/
int
DEFUN(sqlo_set_prefetch_rows, (sth, nrows), 
      sqlo_stmt_handle_t sth AND
      unsigned int nrows)
{
  sqlo_stmt_ptr_t stp;
  int status;

  CHECK_STHANDLE(stp, sth, "sqlo_set_prefetch_rows", SQLO_INVALID_STMT_HANDLE);

  status = _set_prefetch_rows(stp, nrows);
  CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_set_prefetch_rows", "");
  return(status);
}


/*-------------------------------------------------------------------------
 * sqlo_server_version
 *-----------------------------------------------------------------------*/
int
DEFUN(sqlo_server_version, (dbh, bufp, buflen),
      sqlo_db_handle_t dbh AND
      char * bufp AND
      unsigned int buflen)
{
  sqlo_db_ptr_t dbp;
  int status;

  CHECK_DBHANDLE(dbp, dbh, "sqlo_server_version", SQLO_INVALID_STMT_HANDLE);

  status = OCIServerVersion(dbp->srvhp,
                            dbp->errhp,
                            (text *) bufp,
                            (ub4) buflen,
                            OCI_HTYPE_SERVER);
  CHECK_OCI_STATUS_RETURN(dbp, status, "sqlo_server_version", "");
  TRACE(3, fprintf(_get_trace_fp(dbp), "sqlo_server_version returns: %s\n", bufp););
  return(status);
}


/*-------------------------------------------------------------------------
 * sqlo_get_ocol_dtype
 *-----------------------------------------------------------------------*/
int 
DEFUN(sqlo_get_ocol_dtype, (sth, pos), sqlo_stmt_handle_t sth AND unsigned int pos)
{
  sqlo_stmt_ptr_t stp;
  OCIParam * paramd; 
  ub2 dtype;
  int status;
  ub4 num_cols;

  CHECK_STHANDLE(stp, sth, "sqlo_get_ocol_dtype", SQLO_INVALID_STMT_HANDLE);
  TRACE(3, fprintf(_get_trace_fp(stp->dbp), 
                   "sqlo_get_ocol_dtype: Getting datatype of col: %u\n", pos););

  if (0 == stp->num_executions && ! (REFCURSOR == stp->cursor_type) ) {
    /* execute to describe the output */
    while (OCI_STILL_EXECUTING == 
           (status = OCIStmtExecute(stp->dbp->svchp, 
                                    stp->stmthp, 
                                    stp->dbp->errhp, 
                                    (ub4) 0, 
                                    (ub4) 0, 
                                    (OCISnapshot *) 0, 
                                    (OCISnapshot *) 0,
                                    (ub4) OCI_DEFAULT))) {
      SQLO_USLEEP;
    }
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_nget_ocol_dtype", 
                            "OCIStmtExecute(DESCRIBE)");
    ++(stp->num_executions);
  }

  status = OCIAttrGet((dvoid*)stp->stmthp,
                      (ub4) OCI_HTYPE_STMT, 
                      (dvoid *)  &num_cols, 
                      (ub4 *) 0, 
                      (ub4) OCI_ATTR_PARAM_COUNT, 
                      stp->dbp->errhp);

  CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_get_ocol_dtype", "OCIAttrGet(NumberOfColumns)");

  if (pos > (unsigned int) num_cols)
    return SQLO_INVALID_COLPOS;


  /* Get parameters of this statement handle */
  status = OCIParamGet(stp->stmthp, 
                       OCI_HTYPE_STMT, 
                       stp->dbp->errhp,
                       (dvoid *)&paramd, 
                       (ub4) pos);

  CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_get_ocol_dtype", "OCIParamGet");

  /* Get datatype */
  status = OCIAttrGet((dvoid*) paramd, (ub4) OCI_DTYPE_PARAM, 
                      (dvoid*) &dtype,(ub4 *) 0, 
                      (ub4) OCI_ATTR_DATA_TYPE, 
                      (OCIError *)stp->dbp->errhp );

  if (OCI_SUCCESS != status) {
    (void)OCIDescriptorFree(paramd, OCI_DTYPE_PARAM);
    
    CHECK_OCI_STATUS_RETURN(stp->dbp, status, "sqlo_get_ocol_dtype", "OCIAttrGet(datatype)");
  }

  TRACE(3, fprintf(_get_trace_fp(stp->dbp), "sqlo_get_ocol_dtype: datatype: %d (%s)\n", 
                   (int) dtype, 
                   _get_data_type_str((int)dtype)););

  status = OCIDescriptorFree(paramd, OCI_DTYPE_PARAM);
  return( (int) dtype);
  
}
      


/*-------------------------------------------------------------------------
 * sqlo_register_int_handler
 *-----------------------------------------------------------------------*/
int 
DEFUN(sqlo_register_int_handler, (handle, signal_handler), 
      int * handle AND sqlo_signal_handler_t signal_handler )
{
#ifdef HAVE_OSNSUI
  int status;
  status = osnsui(handle, signal_handler, (char *) NULL);
  return status;
#else
  return (SQLO_ERROR);
#endif
}



/*-------------------------------------------------------------------------
 * sqlo_clear_int_handler
 *-----------------------------------------------------------------------*/
int 
DEFUN(sqlo_clear_int_handler, (handle), int handle )
{
#ifdef HAVE_OSNSUI
  int status;
  status = osncui(handle);
  return status;
#else
  return (SQLO_ERROR);
#endif
}

